//-----------------------------------------------------------------------------
//                            CarthaGene
//
// Martin Bouchez, Patrick Chabrier, Christine Gaspin and Thomas Schiex. 
// Unite de Biometrie et Intelligence Artificielle, Toulouse, France
//
// Copyright 1997 INRA (Institut National de la Recherche Agronomique).
// This file is distributed under the terms of the Q Public License version 1.0.
// 
// $Id: BJM_OR.cc,v 1.18 2002/10/08 09:43:41 tschiex Exp $
//
// Description : BJM_OR.
// Divers :
//-----------------------------------------------------------------------------

#include "BJM_OR.h"

#include "Genetic.h"

#include <stdio.h>
#include <string.h>
#include <math.h>

//-----------------------------------------------------------------------------
// Constructeur.
//-----------------------------------------------------------------------------

BJM_OR::BJM_OR() : BioJeuMerged()
{

}

//-----------------------------------------------------------------------------
// Constructeur.
//-----------------------------------------------------------------------------
// Paramtres :
// - accs  l'ensemble des informations du systme. 
// - le numro du jeu
// - type de jeu de donne(Mor)
// - nm nombre de marqueurs.
// - bitjeu champ de bit du jeu de donnes.
// - gauche jeu de donnes.
// - droite jeu de donnes. 
//-----------------------------------------------------------------------------

BJM_OR::BJM_OR(CartaGenePtr cartag,
	       int id,
	       CrossType cross,
	       int nm, 
	       int bitjeu,
	       BioJeu *gauche, 
	       BioJeu *droite) 
  : BioJeuMerged(cartag,
		 id,
 		 cross,
 		 nm, 
 		 bitjeu,
 		 gauche, 
 		 droite)
{
  int i = 0, k = 0;
  int j = 0, l = 0, nbcm = 0;
  int *indmarqpro = new int[Cartage->NbMarqueur + 1];

  for (i = 0; i <= Cartage->NbMarqueur; i++) indmarqpro[i] = 0;

  // recherche des couples communs au deux jeux
  i = 0;
  while (i < NbMarqueur) 
    {
      if (Cartage->BitJeuMarq[k] & BitJeu) 
	{
	  i++;
	  j = i + 1;
	  l = k + 1;
	  while (j <= NbMarqueur) 
	    {
	      if (Cartage->BitJeuMarq[l] & BitJeu) 
		{
		  if (BJgauche->Couplex(k,l) &&
		      BJdroite->Couplex(k,l))
		    {		    
		      if (!indmarqpro[k]) indmarqpro[k] = ++nbcm;
		      if(!indmarqpro[l]) indmarqpro[l] = ++nbcm;
		    }
		  
		  j++;
		}
	      l++;
	    }
	}	
      k++;
    } 

  // allocations

  IndMarq = new int[l];
  
  TwoPointsLOD = new doublePtr[nbcm + 1];
  
  for (i = 0; i <= nbcm; i++) 
    {
      TwoPointsLOD[i]    = new double[nbcm + 1];
      
      for (int j = 0; j <= nbcm; j++) 
	{
	  TwoPointsLOD[i][j] = 0.0;
	}
    }
  
  // remplissage

  TailleCharn = nbcm;

  for (i = 0; i < l; i++) IndMarq[i] = indmarqpro[i];

  delete indmarqpro;

  ComputeTwoPoints();
}

//-----------------------------------------------------------------------------
// Destructeur
//-----------------------------------------------------------------------------

BJM_OR::~BJM_OR()
{
}

//-----------------------------------------------------------------------------
//  Calcul de la vraisemblance maximum par EM.
//-----------------------------------------------------------------------------
// Paramtres : 
// - une carte
// - seuil de convergence
// Valeur de retour : le maximum de vraisemblance
//-----------------------------------------------------------------------------

double BJM_OR::ComputeEM(Carte *map)
{
  int i, j, k;
  int nbmg = 0, nbmd = 0;
  int *vmg, *vmd;
  Carte *cg, *cd;
  double coutg, coutd;

  NbEMCall++;

  // dcompte  gauche, dcompte  droite
  for (i = 0; i < map->NbMarqueur; i++)
    {
      if (Cartage->BitJeuMarq[map->ordre[i]] &
	  BJgauche->BitJeu) nbmg++;
      if (Cartage->BitJeuMarq[map->ordre[i]] &
	  BJdroite->BitJeu) nbmd++;
    }

  vmg = new int[nbmg];
  vmd = new int[nbmd];

  // remplissage  gauche, remplissage  droite
  j = 0;
  k = 0;
  for (i = 0; i < map->NbMarqueur; i++)
    {
      if (Cartage->BitJeuMarq[map->ordre[i]] & 
	  BJgauche->BitJeu) vmg[j++] = map->ordre[i];
      if (Cartage->BitJeuMarq[map->ordre[i]] & 
	  BJdroite->BitJeu) vmd[k++] = map->ordre[i];
    }

  // cration des cartes
  cg = new Carte(Cartage, nbmg, vmg);
  cd = new Carte(Cartage, nbmd, vmd);
  
  // calcul   

  if (nbmg == 0)
    coutg = 0.0;
  else 
    coutg = BJgauche->ComputeEM(cg);
  
  if (nbmd == 0)
    coutd = 0.0;
  else 
    coutd =  BJdroite->ComputeEM(cd);
  
  map->coutEM = coutg + coutd;

  // nettoyage
  delete vmg;
  delete vmd;
  delete cd;
  delete cg;

  return  map->coutEM;
}

//-----------------------------------------------------------------------------
//  Calcul de la vraisemblance maximum par EM.
//  Une interface plus sophistiquee:
//  On converge jusqu'a epsilon1. Si la vraisemblance est inferieure a
//  threshold on laisse tomber, sinon on converge a epsilon2
//-----------------------------------------------------------------------------
// Paramtres : 
// - une carte
// - niveau a partir duquel on convergera a epsilon2
// Valeur de retour : la vraisemblance maximum.
//-----------------------------------------------------------------------------

double BJM_OR::ComputeEMS(Carte *map, 
			  double threshold)
{
  int i, j, k;
  int nbmg = 0, nbmd = 0;
  int *vmg, *vmd;
  Carte *cg, *cd;
  double coutg, coutd;

  NbEMCall++;

  // dcompte  gauche, dcompte  droite
  for (i = 0; i < map->NbMarqueur; i++)
    {
      if (Cartage->BitJeuMarq[map->ordre[i]] &
	  BJgauche->BitJeu) nbmg++;
      if (Cartage->BitJeuMarq[map->ordre[i]] &
	  BJdroite->BitJeu) nbmd++;
    }

  vmg = new int[nbmg];
  vmd = new int[nbmd];

  // remplissage  gauche, remplissage  droite
  j = 0;
  k = 0;
  for (i = 0; i < map->NbMarqueur; i++)
    {
      if (Cartage->BitJeuMarq[map->ordre[i]] & 
	  BJgauche->BitJeu) vmg[j++] = map->ordre[i];
      if (Cartage->BitJeuMarq[map->ordre[i]] & 
	  BJdroite->BitJeu) vmd[k++] = map->ordre[i];    
    }

  // cration des cartes
  cg = new Carte(Cartage, nbmg, vmg);
  cd = new Carte(Cartage, nbmd, vmd);

  // calcul   
  
  if (nbmg == 0)
    coutg = 0.0;
  else 
    coutg = BJgauche->ComputeEMS(cg, threshold);
  
  if (nbmd == 0)
    coutd = 0.0;
  else 
    coutd =  BJdroite->ComputeEMS(cd, threshold);
  
  map->coutEM = coutg + coutd;

  // nettoyage
  delete vmg;
  delete vmd;
  delete cd;
  delete cg;

  return  map->coutEM;
}
//-----------------------------------------------------------------------------
// Calcule fraction de rec. et
// LOD scores pour deux marqueurs
//-----------------------------------------------------------------------------
// Paramtres : 
// - les deux numeros de marqueurs, on doit avoir m1< m2
// - epsilon
// - la fraction de reconbinant(i,o)
// Valeur de retour : le lod
//-----------------------------------------------------------------------------

double BJM_OR::ComputeOneTwoPoints(int m1, 
					 int m2, 
					 double epsilon,
					 double *fr) const
{
  return BJgauche->GetTwoPointsLOD(m1, m2)+
		    BJdroite->GetTwoPointsLOD(m1, m2);
}

//-----------------------------------------------------------------------------
// Remplit les tableaux
//-----------------------------------------------------------------------------

void BJM_OR::ComputeTwoPoints(void)
{
  double lod;
  int i = 0, k = 0;
  int j = 0, l = 0;
  
  while (i < TailleCharn) 
    {
      if (IndMarq[k]) 
	{
	  i++;
	  j = i + 1;
	  l = k + 1;
	  while (j <= TailleCharn) 
	    {
	      if (IndMarq[l]) 
		{
		  lod = BJgauche->GetTwoPointsLOD(k, l)+
		    BJdroite->GetTwoPointsLOD(k, l);
		  TwoPointsLOD[IndMarq[k]][IndMarq[l]] = TwoPointsLOD[IndMarq[l]][IndMarq[k]] = lod;
		  j++;
		}
	      l++;
	    }
	}	
      k++;
    }   
}

//-----------------------------------------------------------------------------
// dtection des groupes, affichage et mise  jour de Cartage.
//-----------------------------------------------------------------------------
// Paramtres : 
// - disthres seuil de la distance 
// - lodthres seuil du lod
// Valeur de retour : le nombre de groupes trouvs
//-----------------------------------------------------------------------------

int BJM_OR::Groupe(double disthres, double lodthres) const
{
  return BioJeu::Groupe(10000.0, lodthres); // 10000.0 = "l'infini"
}


//-----------------------------------------------------------------------------
// Affichage simple d'une carte
//-----------------------------------------------------------------------------
// Paramtres : 
// - la carte
// -
//-----------------------------------------------------------------------------

void BJM_OR::PrintMap(Carte *data) const
{
  BJgauche->PrintMap(data);
  BJdroite->PrintMap(data);
}

//-----------------------------------------------------------------------------
// renvoie d'une carte sous forme de liste
//-----------------------------------------------------------------------------
// Paramtres :
// - l'unit 
// - la carte
//-----------------------------------------------------------------------------

char *** BJM_OR::GetMap(char unit[2], Carte *data) const
{  
  int i, j, k;
  int nbmg = 0, nbmd = 0;
  int *vmg, *vmd;
  Carte *cg, *cd;
  
  // dcompte  gauche, dcompte  droite
  for (i = 0; i < data->NbMarqueur; i++)
    {
      if (Cartage->BitJeuMarq[data->ordre[i]] &
	  BJgauche->BitJeu) nbmg++;
      if (Cartage->BitJeuMarq[data->ordre[i]] &
	  BJdroite->BitJeu) nbmd++;
    }
  
  vmg = new int[nbmg];
  vmd = new int[nbmd];
  
  // remplissage  gauche, remplissage  droite
  j = 0;
  k = 0;
  for (i = 0; i < data->NbMarqueur; i++)
    {
      if (Cartage->BitJeuMarq[data->ordre[i]] & 
	  BJgauche->BitJeu) vmg[j++] = data->ordre[i];
      if (Cartage->BitJeuMarq[data->ordre[i]] & 
	  BJdroite->BitJeu) vmd[k++] = data->ordre[i];
    }
  
  // cration des cartes
  cg = new Carte(Cartage, nbmg, vmg);
  cd = new Carte(Cartage, nbmd, vmd);

  if (nbmg == 0)
    cg->coutEM = 0.0;
  else 
    BJgauche->ComputeEM(cg);
  
  if (nbmd == 0)
    cd->coutEM = 0.0;
  else 
    BJdroite->ComputeEM(cd);

  char*** lmg = BJgauche->GetMap(unit,cg);
  char*** lmd = BJdroite->GetMap(unit,cd);
  i = 0;
  j = 0;
  k = 0;

  // dcompte des cartes

  while (lmg[i++] != NULL);
  while (lmd[j++] != NULL);

  char*** lm =  new char**[i + j - 1];

  lm[i + j - 2] = NULL;

  // fusion

  i = 0;
  j = 0;
  while (lmg[i] != NULL)
    lm[k++] = lmg[i++];
  while (lmd[j] != NULL)
    lm[k++] = lmd[j++];

  // nettoyage
  delete []vmg;
  delete []vmd;
  delete cd;
  delete cg;
  
  delete [] lmg;
  delete [] lmd;
  
  return (lm);
}
//-----------------------------------------------------------------------------
// Affichage dtaill d'une carte(mme principe que computeEM)
//-----------------------------------------------------------------------------
// Paramtres : 
// - la carte
// - le sens
// - la carte globale
//-----------------------------------------------------------------------------

void BJM_OR::PrintDMap(Carte *data, int envers, Carte *dataref)
{
  int i, j, k;
  int nbmg = 0, nbmd = 0;
  int *vmg, *vmd;
  Carte *cg, *cd;

  // dcompte  gauche, dcompte  droite
  for (i = 0; i < data->NbMarqueur; i++)
    {
      if (Cartage->BitJeuMarq[data->ordre[i]] &
	  BJgauche->BitJeu) nbmg++;
      if (Cartage->BitJeuMarq[data->ordre[i]] &
	  BJdroite->BitJeu) nbmd++;
    }

  vmg = new int[nbmg];
  vmd = new int[nbmd];

  // remplissage  gauche, remplissage  droite
  j = 0;
  k = 0;
  for (i = 0; i < data->NbMarqueur; i++)
    {
      if (Cartage->BitJeuMarq[data->ordre[i]] & 
	  BJgauche->BitJeu) vmg[j++] = data->ordre[i];
      if (Cartage->BitJeuMarq[data->ordre[i]] & 
	  BJdroite->BitJeu) vmd[k++] = data->ordre[i];
    }

  // cration des cartes
  cg = new Carte(Cartage, nbmg, vmg);
  cd = new Carte(Cartage, nbmd, vmd);

  if (nbmg != 0) 
    {
      BJgauche->ComputeEM(cg);
      BJgauche->PrintDMap(cg, envers, dataref);
    }
  
  if (nbmd != 0)
    {   
      BJdroite->ComputeEM(cd);
      BJdroite->PrintDMap(cd, envers, dataref);
    }

  // nettoyage
  delete vmg;
  delete vmd;
  delete cd;
  delete cg;

}

//-----------------------------------------------------------------------------
// Accs  une valeur TwoPoints
//-----------------------------------------------------------------------------
// Paramtres : 
// - les numro du couple
// - 
// Valeur de retour : le LOD 2 points
//----------------------------------------------------------------------------

double BJM_OR::GetTwoPointsLOD(int m1, int m2) const
{
  
  return (BJgauche->GetTwoPointsLOD(m1, m2)+
	  BJdroite->GetTwoPointsLOD(m1, m2));

}

//-----------------------------------------------------------------------------
// Accs  une valeur TwoPoints
//-----------------------------------------------------------------------------
// Paramtres : 
// - les rfrences du couple
// - 
// Valeur de retour : la distance 2 points
//----------------------------------------------------------------------------

double BJM_OR::GetTwoPointsDH(int m1, int m2) const
{

  if (BJgauche->Couplex(m1,m2) && BJdroite->Couplex(m1,m2))
    if (HasRH())
    {return (BJgauche->GetTwoPointsDH(m1, m2)); }
    else
    {return (BJgauche->GetTwoPointsDH(m1, m2)+BJdroite->GetTwoPointsDH(m1, m2))/2.0; }

  if (BJgauche->Couplex(m1,m2))
      return BJgauche->GetTwoPointsDH(m1, m2);

  if  (BJdroite->Couplex(m1,m2))
    return BJdroite->GetTwoPointsDH(m1, m2);
  // inherit the large distance of totally non informative pairs
  
  if (HasRH())
  {return (BJgauche->GetTwoPointsDH(m1, m2)); }
  else
  {return (BJgauche->GetTwoPointsDH(m1, m2)+BJdroite->GetTwoPointsDH(m1, m2))/2.0; }
  //return (BJgauche->GetTwoPointsDH(m1, m2)+BJdroite->GetTwoPointsDH(m1, m2))/2.0;
}
