//-----------------------------------------------------------------------------
//                            CarthaGene
//
// Martin Bouchez, Patrick Chabrier, Christine Gaspin and Thomas Schiex. 
// Unite de Biometrie et Intelligence Artificielle, Toulouse, France
//
// Copyright 1997 INRA (Institut National de la Recherche Agronomique).
// This file is distributed under the terms of the Q Public License version 1.0.
// 
// $Id: BJS_CO.h,v 1.15 2003/02/20 13:46:37 chabrier Exp $
//
// Description : Prototype de la classe BJS_CO.
// Divers : 
//-----------------------------------------------------------------------------

#ifndef _BJS_CO_H
#define _BJS_CO_H

#include "BioJeuSingle.h"

#include "Constraints.h"

/** Classe des jeux de donnes Contraintes*/
class BJS_CO : public BioJeuSingle {
 public:

  /** Liste des contraintes.
  */
  Constraint *Constraints;

  /** Constructeur. */
  BJS_CO();

  /** Constructeur.
      @param cartag accs  l'ensemble des informations du systme. 
      @param id le numro du jeu.
      @param cross type de jeu de donne.
      @param nm nombre de marqueurs.
      @param taille le nombre de contraintes
      @param bitjeu champ de bit du jeu de donnes.
      @param cons la liste des contraintes
  */
  BJS_CO(CartaGenePtr cartag,
	 int id,
	 charPtr nomjeu,
	 CrossType cross,	 
	 int nm, 
	 int taille,
	 int bitjeu,
	 Constraint *cons);
  
  /** Destructeur. */
  ~BJS_CO();

  /** Retourne un champ des jeux fusionns sur l'ordre(!=CON).
      @return un champ de bits
  */
  inline int BJS_CO::GetBJMO() const
    {
      return 0;
    };

  /** Compatibilite entre deux echantillons d'un marqueur
      @param numarq1 le numero du marqueur 1
      @param numarq2 le numero du marqueur 2
  */
  int Compatible(int numarq1, int numarq2) const;

  /** Fusion de deux marqueurs pour la creation d'un haplotype. Le
      marqueur 1 est utilise pour stocker le resultat.
      @param numarq1 marqueur 1 
      @param numarq2 marqueur 2 
      @return nothing.  
  */
  void Merge(int marq1,int marq2) const;


  /** Affichage des contraintes*/
  void DumpEch(void) const;

  
  /** Test l'existance d'un couple(pas de sens)
      @param m1 marqueur 1
      @param m2 marqueur 2
      @return 0
  */
  int Couplex(int m1, int m2) const;

  
  /** mthode qui ne fait rien pour la cohrence!
    */
  void ComputeTwoPoints(void) {};

  /** mthode qui ne fait rien pour la cohrence!
  */
  virtual void DumpEchMarq(int numarq) const {};
  
  /** mthode qui ne fait rien pour la cohrence!
  */
  double GetTwoPointsLOD(int m1, int m2) const {return 0.0;};

  /** mthode qui ne fait rien pour la cohrence!
    */
  virtual void DumpTwoPointsLOD() const {};

    /** Calcul du maximum de vraisemblance par EM.     
      @param data la carte
      @return le maximum de vraissemblance
  */
  double ComputeEM(Carte *data);  

  /** Calcul de la vraisemblance maximum par EM.
      Une interface plus sophistique du calcul de la vraisemblance par EM:
      On converge jusqu' epsilon1. Si la vraisemblance est infrieure 
      threshold on laisse tomber, sinon on converge  epsilon2.
      @param data la carte
      @param threshold niveau a partir duquel on convergera a epsilon2
      @return le maximum de vraissemblance
  */
  double ComputeEMS (Carte *data, 
		     double threshold);
  
  /** Affichage dtaill d'une carte
      @param data la carte
      @param envers  l'envers ou  l'endroit
  */
  void PrintDMap(Carte *data, int envers, Carte *dataref);

  /** Matrice 2p des distances 5bidon)
      @param unit l'unit de distance: 
      'k' = kosambi ou centiray, 
      'h' = haldane ou centiray.
  */
  inline void PrintTwoPointsDist(char unit[2]) const {};
  
 private:

  /** passage  priv d'une virtuelle hrite = choix d'implmentation. */
  virtual double LogInd(int m1, int m2, int *nbdata) const {
    sprintf(bouf,"Bug, mthode interdite = LogInd!\n");
    pout(bouf);
    return 0.0;
  };

  /** passage  priv d'une virtuelle hrite = choix d'implmentation. */
  virtual double EspRec(int m1, int m2, double theta, double *loglike) const {
    sprintf(bouf,"Bug, mthode interdite = EspRec!\n");
    pout(bouf);
    return 0.0; 
  };

  /** passage  priv d'une virtuelle hrite = choix d'implmentation. */
  double GetTwoPointsDH(int m1, int m2) const {
    sprintf(bouf,"Bug, mthode interdite = GetTwoPointsDH!\n");
    pout(bouf);
    return 0.0; 
  };
  
  /** passage  priv d'une virtuelle hrite = choix d'implmentation. */
  double GetTwoPointsFR(int m1, int m2) const {
    return 0.0;
  };
  
  /** passage  priv d'une virtuelle hrite = choix d'implmentation. */
  double ComputeExpected(const Carte *data, double *expected) {
    sprintf(bouf,"Bug, mthode interdite = ComputeExpected!\n");
    pout(bouf);
    return 0.0;
  };

  /** Calcul stat. suffisantes pour le 2pt
      @param m1 premier marqueur
      @param m2 second marqueur
      @param ss statistiques suffisantes (n[1..3])
  */
  void Prepare2pt(int m1, int m2,  int *ss) const {
    sprintf(bouf,"Bug, mthode interdite = Prepare2pt!\n");
    pout(bouf);
    return;
  };

  /** Calcul des estimateurs 2pt de vrais. max
      @param par vecteur des parametres du modele (theta,r mis a jour)
      @param ss statistiques suffisantes (n[1..3])
  */
  void Estimate2pt(double *par, int *ss) const  {
    sprintf(bouf,"Bug, mthode interdite = Estimate2pt!\n");
    pout(bouf);
    return;
  };

  /** Calcul de la vraisemblance 2pt etant donnes les parametres du
      modele.
      @param par vecteur des parametres du modele (theta,r)
      @param ss statistiques suffisantes (n[1..3])
  */
  double LogLike2pt(double *par, int *ss) const  {
    sprintf(bouf,"Bug, mthode interdite = LogLike2pt!\n");
    pout(bouf);
    return 0.0;
  };

  /** passage  priv d'une virtuelle hrite = choix d'implmentation. */
  virtual double ComputeOneTwoPoints(int m1, int m2, double epsilon, double *fr) const {
    sprintf(bouf,"Bug, mthode interdite = ComputeOneTwoPoints!\n");
    pout(bouf);
    return 0.0;
  };

  /** passage  priv d'une virtuelle hrite = choix d'implmentation. */
  void PreparEM(const Carte *data) {
    sprintf(bouf,"Bug, mthode interdite = PreparEM!\n");
    pout(bouf);
  };
  
  /** passage  priv d'une virtuelle hrite = choix d'implmentation. */
  void NetEM(const Carte *data) {
    sprintf(bouf,"Bug, mthode interdite = NetEM!\n");
    pout(bouf);
  };
  
  /** passage  priv  = choix d'implmentation. */  
  int Maximize(Carte *data, double *expected) const {
    sprintf(bouf,"Bug, mthode interdite = Maximize!\n");
    pout(bouf);
    return -1;
  };

};

#endif /* _BJS_CO_H */
