//-----------------------------------------------------------------------------
//                            CarthaGene
//
// Martin Bouchez, Patrick Chabrier, Christine Gaspin and Thomas Schiex. 
// Unite de Biometrie et Intelligence Artificielle, Toulouse, France
//
// Copyright 1997 INRA (Institut National de la Recherche Agronomique).
// This file is distributed under the terms of the Q Public License version 1.0.
// 
// $Id: BioJeuSingle.cc,v 1.17 2003/01/09 10:53:59 chabrier Exp $
//
// Description : Classe de base BioJeuSingle.
// Divers : 
//-----------------------------------------------------------------------------

#include "BioJeuSingle.h"

#include "Genetic.h"

#include <stdio.h>
#include <string.h>
#include <math.h>

//-----------------------------------------------------------------------------
// Constructeur.
//-----------------------------------------------------------------------------

BioJeuSingle::BioJeuSingle()
{
  Cartage = NULL;
  NomJeu = NULL;
  Cross = Unk;
  BitJeu = 0;
  NbMarqueur = 0;
  TailleEchant = 0;
  Echantillon = NULL; 
  TwoPointsFR = NULL; 
  TwoPointsDH = NULL; 
  TwoPointsLOD = NULL; 

}


//-----------------------------------------------------------------------------
// Constructeur.
//-----------------------------------------------------------------------------
// Paramtres :
// - le rfrentiel
// - le numro
// - le chemin et nom du fichier de donnees
// - le type
// - le nombre de marqueurs
// - la taille de l'echantillon
// - le champ de bit
// - la matrice de l'chantillon
// Valeur de retour : 
//-----------------------------------------------------------------------------

BioJeuSingle::BioJeuSingle(CartaGenePtr cartag,
			   int id,
			   charPtr nomjeu,
			   CrossType cross,
			   int nm, 
			   int taille,
			   int bitjeu,
			   int *indmarq,
			   Obs **echantil)
  : BioJeu(cartag, id, cross, nm, bitjeu)
{  

  Em_Max_Theta = EM_MAX_THETA;
  Em_Min_Theta = EM_MIN_THETA;

  NomJeu = new char[strlen(nomjeu)+1];
  strcpy(NomJeu, nomjeu);

  TailleEchant = taille;
  IndMarq = indmarq;
  Echantillon = echantil;

  TwoPointsFR = new doublePtr[NbMarqueur + 1];
  TwoPointsDH = new doublePtr[NbMarqueur + 1];
  TwoPointsLOD = new doublePtr[NbMarqueur + 1];

  for (int i = 0; i <= NbMarqueur; i++) 
    {
      TwoPointsFR[i]    = new double[NbMarqueur + 1];
      TwoPointsDH[i]    = new double[NbMarqueur + 1];
      TwoPointsLOD[i]    = new double[NbMarqueur + 1];

      for (int j = 0; j <= NbMarqueur; j++) 
	{
	  TwoPointsFR[i][j] = 0.0;
	  TwoPointsDH[i][j] = 0.0;
	  TwoPointsLOD[i][j] = 0.0;
	}
    }

}

//-----------------------------------------------------------------------------
// Destructeur
//-----------------------------------------------------------------------------

BioJeuSingle::~BioJeuSingle()
{  

  // verrue pour Con

  if (Echantillon) {

    for (int i = 0; i <= NbMarqueur; i++)
      {
	delete [] TwoPointsDH[i];
	delete [] TwoPointsFR[i];
	delete [] TwoPointsLOD[i];
	delete [] Echantillon[i];
      }
    delete [] TwoPointsDH;
    delete [] TwoPointsFR;
    delete [] TwoPointsLOD;
    delete [] Echantillon;
    delete IndMarq;
  }
  
  delete NomJeu;
}


//-----------------------------------------------------------------------------
// Remplit les tableaux
//-----------------------------------------------------------------------------
// Paramtres : 
// - 
// - 
// Valeur de retour : 
//-----------------------------------------------------------------------------

void BioJeuSingle::ComputeTwoPoints(void)
{
  
  double fr;
  double lod;
  int i = 0, k = 0;
  int j = 0, l = 0;
  
  while (i < NbMarqueur) 
    {
      if (IndMarq[k]) 
	{
	  i++;
	  j = i + 1;
	  l = k + 1;
	  while (j <= NbMarqueur) 
	    {
	      if (IndMarq[l]) 
		{
		  lod = ComputeOneTwoPoints(k, l, Epsilon2, &fr);
		  TwoPointsLOD[IndMarq[k]][IndMarq[l]] = 
		    TwoPointsLOD[IndMarq[l]][IndMarq[k]] = lod;
		  TwoPointsFR[IndMarq[k]][IndMarq[l]] = 
		    TwoPointsFR[IndMarq[l]][IndMarq[k]] = fr;
		  TwoPointsDH[IndMarq[k]][IndMarq[l]] = 
		    TwoPointsDH[IndMarq[l]][IndMarq[k]] = 
		    (HasRH() ? Theta2Ray(TwoPointsFR[j][i]) : Haldane(TwoPointsFR[j][i]));
		  j++;
		}
	      l++;
	    }
	}	
      k++;
    }   
}

//-----------------------------------------------------------------------------
// Affichage des donnes
//-----------------------------------------------------------------------------
// Paramtres : 
// - 
// - 
// Valeur de retour : 
//----------------------------------------------------------------------------

void BioJeuSingle::DumpEch(void) const
{
  int i = 0, j, k = 0;

  sprintf(bouf, "%10s : Individuals...\n","Markers");
  pout(bouf);
  // Nom des marqueurs et les observations.
  while (i < NbMarqueur) 
    {
      if (IndMarq[k]) {
	i++;
	sprintf(bouf, "%10s : ", Cartage->NomMarq[k]);
	pout(bouf);
	for (j=1; j<=TailleEchant; j++)

	  sprintf(bouf, "%x", Echantillon[IndMarq[k]][j]);
	pout(bouf);

	sprintf(bouf,  "\n");
	pout(bouf);
      }
      k++;
    }
}

//-----------------------------------------------------------------------------
// Affichage des donnes pour un marqueur
//-----------------------------------------------------------------------------
// Paramtres : 
// - le numro d'un marqueur
// - 
// Valeur de retour : 
//----------------------------------------------------------------------------
void BioJeuSingle::DumpEchMarq(int numarq) const
{
 
  for (int i=1; i <= TailleEchant; i++) 
    {
      sprintf(bouf, "%x", GetEch(numarq, i));
      pout(bouf);
      fflush(stdout);
    }
}

//-----------------------------------------------------------------------------
// Accs  une valeur TwoPoints
//-----------------------------------------------------------------------------
// Paramtres : 
// - les numro du couple
// - 
// Valeur de retour : le LOD 2 points
//----------------------------------------------------------------------------
   
double BioJeuSingle::GetTwoPointsLOD(int m1, int m2) const
{
  return TwoPointsLOD[IndMarq[m1 * ((Cartage->BitJeuMarq[m1] & BitJeu) > 0)]][IndMarq[m2 * ((Cartage->BitJeuMarq[m2] & BitJeu) > 0)]];
}

//-----------------------------------------------------------------------------
// Accs  une valeur TwoPoints
//-----------------------------------------------------------------------------
// Paramtres : 
// - les numro du couple
// - 
// Valeur de retour : la distance 2 points
//----------------------------------------------------------------------------
   
double BioJeuSingle::GetTwoPointsDH(int m1, int m2) const
{
  return  (HasRH() ? Theta2Ray(TwoPointsFR[IndMarq[m1 * ((Cartage->BitJeuMarq[m1] & BitJeu) > 0)]][IndMarq[m2 * ((Cartage->BitJeuMarq[m2] & BitJeu) > 0)]]) : Haldane(TwoPointsFR[IndMarq[m1 * ((Cartage->BitJeuMarq[m1] & BitJeu) > 0)]][IndMarq[m2 * ((Cartage->BitJeuMarq[m2] & BitJeu) > 0)]]));
}

//-----------------------------------------------------------------------------
// Accs  une valeur TwoPoints
//-----------------------------------------------------------------------------
// Paramtres : 
// - les numro du couple
// - 
// Valeur de retour : la Fraction de recombinaison 2 points
//----------------------------------------------------------------------------
   
double BioJeuSingle::GetTwoPointsFR(int m1, int m2) const
{
  return TwoPointsFR[IndMarq[m1 * ((Cartage->BitJeuMarq[m1] & BitJeu) > 0)]][IndMarq[m2 * ((Cartage->BitJeuMarq[m2] & BitJeu) > 0)]];
}


//-----------------------------------------------------------------------------
// Affichage de la matrice TwoPoints Lod
//-----------------------------------------------------------------------------
// Paramtres : 
// - 
// Valeur de retour :
//----------------------------------------------------------------------------
   
void BioJeuSingle::DumpTwoPointsLOD() const
{
  int i = 0, k = 0;
  int j = 0, l = 0;

  while (i < NbMarqueur) 
    {
      if (IndMarq[k]) 
	{
	  i++;
	  sprintf(bouf, "%3d :", k);
	  pout(bouf);
	  j = 0;
	  l = 0;
	  while (j < NbMarqueur) 
	    {
	      if (IndMarq[l]) 
		{
		  j++;
		  sprintf(bouf, "%5.2f ", GetTwoPointsLOD(k, l));
		  pout(bouf);
		}
	      l++;
	    }
	  sprintf(bouf,  "\n");
	  pout(bouf);
	}	
      k++;
    }   
}

void BioJeuSingle::DumpTwoPointsFR() const
{
  int i = 0, k = 0;
  int j = 0, l = 0;

  while (i < NbMarqueur) 
    {
      if (IndMarq[k]) 
	{
	  i++;
	  sprintf(bouf, "%3d :", k);
	  pout(bouf);
	  j = 0;
	  l = 0;
	  while (j < NbMarqueur) 
	    {
	      if (IndMarq[l]) 
		{
		  j++;
		  sprintf(bouf, "%5.2f ", GetTwoPointsFR(k, l));
		  pout(bouf);
		}
	      l++;
	    }
	  sprintf(bouf,  "\n");
	  pout(bouf);
	}	
      k++;
    }   
}
