//-----------------------------------------------------------------------------
//                            CarthaGene
//
// Martin Bouchez, Patrick Chabrier, Christine Gaspin and Thomas Schiex. 
// Unite de Biometrie et Intelligence Artificielle, Toulouse, France
//
// Copyright 1997 INRA (Institut National de la Recherche Agronomique).
// This file is distributed under the terms of the Q Public License version 1.0.
// 
// $Id: CartaGene.cc,v 1.84 2003/09/10 07:29:01 tschiex Exp $
//
// Description : Classe API du module.
// Divers : 
//-----------------------------------------------------------------------------


#include "CartaGene.h"
#include "BJS_BC.h"
#include "BJS_IC.h"
#include "BJS_RH.h"
#include "BJS_RHD.h"
#include "BJM_GE.h"
#include "BJM_OR.h"
#include "BJS_CO.h"
#include "BioJeu.h"
#include "Constraints.h"
#include "AlgoGen.h"

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <math.h>
#ifdef __WIN32__
#include <winsock2.h>
#endif


//-----------------------------------------------------------------------------
// Constructeur.
//-----------------------------------------------------------------------------
CartaGene::CartaGene()
{

  Fout = NULL;
  NbMarqueur = 0;
  NbJeu = 0;
  NbIndividu = 0;
  NomMarq = NULL;
  BitJeuMarq = NULL;
  Jeu = NULL;
  ArbreJeu = NULL;
  Group = NULL;
  DisThres = 0.0;
  Merged = NULL;
  Represents = NULL;
  LODThres = 0.0;
  Robustness = 1e100;
  VerboseFlag = 0;
  QuietFlag = 0;
  NbMS = 0;
  MarkSelect = NULL;
  wordre = NULL;

  StopFlag = 0;

  Heap.Init(this);
}
//-----------------------------------------------------------------------------
// Destructeur.
//-----------------------------------------------------------------------------

CartaGene::~CartaGene()
{
  int i;
  NodintPtr padj, padjw;
  NodptrPtr plink; 

  if (Fout != NULL)
    fclose(Fout);


  for (i = 1; i <= NbMarqueur; i++) {
    delete NomMarq[i];
  }
  delete NomMarq;
  delete Merged;
  delete MarkSelect;
  delete Represents;

  for (i = 1; i <= NbJeu; i++) delete Jeu[i];
  delete Jeu;
  delete BitJeuMarq;

  plink = Group;

  while (plink != NULL) {

    padj = plink->first;

    while (padj != NULL) 
      {	
	padjw = padj;
	padj = padj->next;
	delete padjw; 
      }

    Group = plink;
    plink = plink->next;
    delete Group;
  } 
}

//-----------------------------------------------------------------------------
// gestion de la trace
//-----------------------------------------------------------------------------

void CartaGene::Tracer(char *fileName)
{
  FILE *ftmp;
  
  // demande de fermeture
  if (strcmp(fileName, "") == 0)
    {
      // un fichier est dja ouvert
      if (Fout != NULL)
	{
	  fclose(Fout);
	  Fout = NULL;
	}
    } else
      {
	// chec de l'ouverture (rien faire)
	ftmp = fopen(fileName,"a");
	if (ftmp == NULL) 
	  {
	    perr("The opening of the file has failed!\n");
	  }
	else 
	  {
	    // un fichier est dja ouvert
	    if (Fout != NULL)
	      {
		fclose(Fout);
		Fout = NULL;
	      }
	    Fout = ftmp;
	  }
      }
}

//-----------------------------------------------------------------------------
// affiche la version.
//-----------------------------------------------------------------------------

void CartaGene::Version(void)
{
  sprintf(bouf,"\n"); pout(bouf);
  sprintf(bouf,"   CarthaGene version 0.99, Copyright (c) 1997-2003 (INRA).\n"); pout(bouf);
  sprintf(bouf,"\n"); pout(bouf);
  sprintf(bouf,"   CarthaGene comes with ABSOLUTELY NO WARRANTY. \n"); pout(bouf);
  sprintf(bouf,"   CarthaGene is free software. You are welcome to redistribute it,\n"); pout(bouf);
  sprintf(bouf,"   under certain conditions. See the ""License"" file for information.\n"); pout(bouf);
  sprintf(bouf,"\n"); pout(bouf);
  sprintf(bouf,"Type 'help' for help.\n"); pout(bouf);
  sprintf(bouf,"\n"); pout(bouf);
}
//-----------------------------------------------------------------------------
// Netoyeur.
//-----------------------------------------------------------------------------

void CartaGene::Clear(void)
{
  int i;
  NodintPtr padj, padjw;
  NodptrPtr plink; 

  for (i = 1; i <= NbMarqueur; i++) {
    delete [] NomMarq[i];

  }
  delete [] NomMarq;
  delete [] BitJeuMarq;
  delete Merged;
  delete Represents;

  for (i = 1; i <= NbJeu; i++) delete Jeu[i];
  delete [] Jeu;

  plink = Group;

  while (plink != NULL) {

    padj = plink->first;

    while (padj != NULL) 
      {	
	padjw = padj;
	padj = padj->next;
	delete padjw; 
      }

    Group = plink;
    plink = plink->next;
    delete Group;
  } 

  delete [] wordre;

  delete [] MarkSelect;

  Merged = NULL;
  Represents = NULL;
  NbMarqueur = 0;
  NbJeu = 0;
  NbIndividu = 0;
  NomMarq = NULL;
  BitJeuMarq = NULL;
  Jeu = NULL;
  ArbreJeu = NULL;
  Group = NULL;
  DisThres = 0.0;
  LODThres = 0.0;
  Robustness = 1e100;
  VerboseFlag = 0;
  QuietFlag = 0;
  NbMS = 0;
  MarkSelect = NULL;
  wordre = NULL;
  
  Heap.Init(this);

}
//-----------------------------------------------------------------------------
// Lecture de la clef de la license
//-----------------------------------------------------------------------------
// Paramtres : la clef
// - 
// - 
// Valeur de retour : aucune
//-----------------------------------------------------------------------------
void CartaGene::License(const char *key)
{
  // License supprimee depuis mise sous QPL
}
//-----------------------------------------------------------------------------
// Lit le type de croisement utilis dans le fichier de donnes
//-----------------------------------------------------------------------------
// Paramtres : le nom du fichier
// - 
// - 
// Valeur de retour : le type de croisement
//-----------------------------------------------------------------------------

CrossType CartaGene::ProbeCrossType(char DataFile[256])
{
  FILE *fdata;
  char line[256];
  char *tok;

  if ((fdata=fopen(DataFile, "r")) == NULL)
    {
      perror(DataFile);
      return(Unk);
    }
  
  fgets(line, 256, fdata);
  tok= line;
  while (*tok) {
    *tok= tolower(*tok);
    tok++;
  }

  fclose(fdata);

  tok = strtok(line, " \t\f\n\r\v");

  if (strcmp(tok,"data") || strcmp(strtok(NULL," \t\f\n\r\v"),"type")) {
    sprintf(bouf, "%s : Bad header on line 1\n", DataFile);
    perr(bouf);
    return(Unk);
  }
  
  tok = strtok(NULL," \t\f\n\r\v");

  if (!strcmp(tok,"f2")) {// backcross ou intercross
      tok = strtok(NULL," \t\f\n\r\v");
      if (!strcmp(tok, "backcross")) return(BC);    
      if (!strcmp(tok, "intercross")) return(IC);
      sprintf(bouf, "%s : Bad header on line 1\n", DataFile);
      perr(bouf);
      return(Unk);
  }

  if (!strcmp(tok,"ri")) {//RIL
      tok = strtok(NULL," \t\f\n\r\v");
      if (!strcmp(tok, "self")) return(RISelf);    
      if (!strcmp(tok, "sib")) return(RISib);
      sprintf(bouf, "%s : Bad header on line 1\n", DataFile);
      perr(bouf);
      return(Unk);
  }

  if (!strcmp(tok,"radiated") && !strcmp(strtok(NULL," \t\f\n\r\v"),"hybrid")) {//RH
      tok = strtok(NULL," \t\f\n\r\v");
      if (!tok)  return(RH);
      if (!strcmp(tok, "diploid")) return(RHD);

      sprintf(bouf, "%s : Bad header on line 1\n", DataFile);
      perr(bouf);
      return(Unk);
  }
  
  if (!strcmp(tok,"constraint")) return(Con);
  
    sprintf(bouf, "%s : Bad header on line 1\n", DataFile);
    perr(bouf);
    return(Unk);
}

//-----------------------------------------------------------------------------
// Charge un jeu de donnes biologiques  partir d'un fichier :
// - mise  jour des informations sur les marqueurs
// - cration et stockage d'un BioJeu
// - mise  jour de la liste des marqueurs actifs.
//-----------------------------------------------------------------------------
// Paramtres : 
// - le nom du fichier
// Valeur de retour : Chaine de car. contenant 
// - le numro du jeu
// - le type du jeu
// - le nombre de marqeurs
// - le nombre d'individus
// ou une chaine vide en cas d'erreur.
//-----------------------------------------------------------------------------

char * CartaGene::CharJeu(char *fileName)
{

  struct stat buf;
  FILE *fdata;

  CrossType Cross;
  int nbi = 0, nbm = 0, nbnm = 0; 
  Obs **Ech;
  int bitjeu;
  BioJeu *bjp;

  int numjeu = NbJeu + 1;
  charPtr *nomarq;
  charPtr *nomarqj;
  charPtr *nomarqfus;
  int *bitjeufus;
  BioJeu **jeufus;
  int *marqex;
  int *indmarq;
  char *mergemarq;
  int *repmarq;

  char *vide;
  char *lineres;
  char c, d, l[256], tr[256], l2[256], l3[256];
  char *line = l;
  int i, j, k, pos, m1, m2, m3;
  double pen;


  // pour retourner une liste vide(voir conventions)
  vide = new char[1];
  vide[0] = '\0';

  // On teste si le nombre maximum de jeu charges est atteint.
  if (NbJeu >= MAXNBJEU) {
    sprintf(bouf, "%s : Maximum number of simultaneously datasets loaded exceeded (%d)\n", fileName,MAXNBJEU);
    perr(bouf);
    return vide;
  }

  Constraint *cons, *con;

  if (stat(fileName, &buf) == -1) 
    {
      perror(fileName);
      return vide;
    }
  
  Cross = ProbeCrossType(fileName);
  
  if ( Cross == Unk ) return vide;

  lineres = new char[256];

  fdata=fopen(fileName, "r"); // no error test needed, done in "ProbeCrossType"
  fgets(lineres, 256, fdata); // stockage pour la rponse

  
  // chargement diffrent...
  if ( Cross != Con )
    {  
      fgets(line, 256, fdata);
      if ((i=sscanf(line, "%d %d%n", &nbi, &nbm, &pos)) < 2) {
	sprintf(bouf, "%s : Syntax error on line 2, number of loci or individual missing\n", fileName);
	perr(bouf);
	delete lineres;
	return vide;
      }
      
      line += pos;
      if (sscanf(line, "%*d%n", &pos) != EOF) {//  more to read ?
	line += pos;
	if (sscanf(line," case%n",&pos)) line += pos;
      }
      
      for (i = 0; i < 256;i++) tr[i] = 0;
      
      if (Cross != IC) {
	tr[' '] = ' ';
	tr['-'] = '-';
	tr['A'] = '0';
	tr['B'] = 'A';
	tr['H'] = 'A';
      }
      else {
	tr[' '] = ' ';
	tr['-'] = '-';
	tr['A'] = 'A';
	tr['B'] = 'B';
	tr['C'] = 'C';
	tr['D'] = 'D';
	tr['H'] = 'H';
	tr['0'] = '0'; // 0000
	tr['1'] = 'A'; // 0001
	tr['2'] = '2'; // 0010
	tr['3'] = '3'; // 0011
	tr['4'] = '4'; // 0100
	tr['5'] = '5'; // 0101
	tr['6'] = 'H'; // 0110
	tr['7'] = 'D'; // 0111
	tr['8'] = 'B'; // 1000
	tr['9'] = '9'; // 1001
	tr['a'] = 'a'; // 1010
	tr['b'] = 'b'; // 1011
	tr['c'] = 'c'; // 1100
	tr['d'] = 'd'; // 1101
	tr['e'] = 'C'; // 1110
	tr['f'] = '-'; // 1111
      }

      if (sscanf(line," %*s%n",&pos) != EOF) {
	line += pos;
	while (1)	  {
	  if ((i = sscanf(line, " %c=%c%n", &c, &d, &pos)) < 2 ) break;
	  line += pos;
	  tr[c] = tr[d];
	}
      }
      
      // Allocation mmoire pour les tableaux provisoires 
      nomarq = new charPtr[nbm + 1];
      nomarq[0] = NULL;
      nomarqj = new charPtr[nbm + 1];
      nomarqj[0] = NULL;
    }

  marqex = new int[NbMarqueur + nbm + 1];
  for (i = 0; i <= NbMarqueur + nbm; i++) marqex[i] = 0;
     
  // bitjeu est le masque bit a bit correspondant au jeu de donnee "numjeu"
  bitjeu = 1<<(numjeu - 1); 
  
  // chargement diffrent...
  if (Cross != Con)
    {

      // Allocation mmoire pour le tableau de donnees.
      Ech = new Obs*[nbm + 1];
      for (i = 0; i <= nbm; i++) 
	{
	  Ech[i] = new Obs[nbi + 1];
	  Ech[i][0] = Obs1111;
	}
      
      for (i = 0; i <= nbi; i++) Ech[0][i] = Obs1111;
      
      for (i = 1; i <= nbm ; i++) 
	{
	  fscanf(fdata, " *%s ", line);


	  // controle de l'unicit au sein d'un jeu

	  // remplissage d'une liste pour le jeu

	  nomarqj[i] = new char[strlen(line)+1];
	  strcpy(nomarqj[i], line);

	  for (k = 1; k <i ; k++) 
	    {
#ifdef __WIN32__
	      if (!_stricmp(line, nomarqj[k])) break;
#else
	      if (!strcasecmp(line, nomarqj[k])) break;
#endif
	    }
	  if ( k < i)
	    {
	      sprintf(bouf, "%s: Syntax error on marker, %s is not uniq.\n", fileName,line);
	      pout(bouf);
	      for (k = 0; k <= nbm ; k++) 
		delete Ech[k];      
	      for (k = 1; k <= nbnm; k++) 
		delete nomarq[k];
	      delete nomarq;
	      for (k = 1; k < i; k++) 
		delete nomarqj[k];
		delete nomarqj;
	      delete marqex;
	      delete lineres;
	      delete Ech;
	      fclose(fdata);
	      return vide;
	    }
	  
	  // recherche du marqueur
	  for (k = 1; k <= NbMarqueur ; k++) 
	    {
#ifdef __WIN32__
	      if (!_stricmp(line, NomMarq[k])) break;
#else
	      if (!strcasecmp(line, NomMarq[k])) break;
#endif
	    }
	  // trouv ou pas
	  if ( k <= NbMarqueur ) 
	    marqex[k] = i;
	  else 
	    {      
	      nomarq[++nbnm] = new char[strlen(line)+1];
	      
	      marqex[NbMarqueur + nbnm] = i;
	      
	      strcpy(nomarq[nbnm], line);
	    }
	  
	  do c = getc(fdata);
	  while (isspace(c));
	  
	  for (j = 1; j <= nbi; j++) {
	    c = tr[c];
	    switch(c) 
	      {
	      case ' ': j--; break;
	      case 'A': Ech[i][j] = Obs0001;   break;
	      case 'H': Ech[i][j] = Obs0110;   break;
	      case 'B': Ech[i][j] = Obs1000;   break;
	      case 'C': Ech[i][j] = Obs1110;   break;
	      case 'D': Ech[i][j] = Obs0111;   break;
	      case '-': Ech[i][j] = Obs1111;   break;
	      case '2': Ech[i][j] = Obs0010;   break;
	      case '3': Ech[i][j] = Obs0011;   break;
	      case '4': Ech[i][j] = Obs0100;   break;
	      case '5': Ech[i][j] = Obs0101;   break;
	      case '9': Ech[i][j] = Obs1001;   break;
	      case 'a': Ech[i][j] = Obs1010;   break;
	      case 'b': Ech[i][j] = Obs1011;   break;
	      case 'c': Ech[i][j] = Obs1100;   break;
	      case 'd': Ech[i][j] = Obs1101;   break;
	      case '0': 
		Ech[i][j] = Obs0000;   
		if (Cross != IC) break;
		
	      default:
		sprintf(bouf, "%s: Syntax error on marker ", fileName);
		pout(bouf);
		sprintf(bouf, "%s, position %d, code '%c'\n", line, j,c);
		pout(bouf);
		for (k = 0; k <= nbm ; k++) 
		  delete Ech[k];      
		for (k = 1; k <= nbnm; k++) 
		  delete nomarq[k];		
		delete nomarq;
		for (k = 1; k < i; k++) 
		  delete nomarqj[k];
		delete nomarqj;
		delete marqex;
		delete lineres;
		delete Ech;
		fclose(fdata);
		return vide;
	      } 
	    c = getc(fdata);
	  }
	}
    } else
      {
	cons = NULL;
	j = 1;
	while (1)
	  {
	    i = fscanf(fdata, "%s %s %s %lf", l, l2, l3, &pen);
	    j++;
	    if (i == EOF)  break; 
	    if (i<4)
	      {
	    	sprintf(bouf, "%s : Syntax error on line number %d\n", 
			fileName, j);
		perr(bouf);
		delete lineres;
		return vide;
	      }
	    m1 = 0;
	    m2 = 0;
	    m3 = 0;
	    // recherche des marqueurs
	    for (k = 1; k <= NbMarqueur ; k++) 
	      {
#ifdef __WIN32__
		if (!_stricmp(l, NomMarq[k])) 
#else
		  if (!strcasecmp(l, NomMarq[k])) 
#endif
		  { 
		    m1 = k;
		    marqex[k] = 1;
		  }
#ifdef __WIN32__
		if (!_stricmp(l2, NomMarq[k])) 
#else
		  if (!strcasecmp(l2, NomMarq[k])) 
#endif
		  {
		    m2 = k;
		    marqex[k] = 1;
		  }
#ifdef __WIN32__
		if (!stricmp(l3, NomMarq[k])) 
#else
		  if (!strcasecmp(l3, NomMarq[k])) 
#endif
		  {
		    m3 = k;
		    marqex[k] = 1;
		  }
		if (m1 != 0 && m2 != 0 && m3 != 0) break;		
	      }
	    // trouv ou pas
	    if ( k > NbMarqueur ) {
	      if (m2 == 0) { 
		sprintf(bouf, "%s : Syntax error on line number %d, the marker %s does not exist.\n", fileName, j, l2);
		perr(bouf);
	      }
	      else if (m3 == 0) {
		sprintf(bouf, "%s : Syntax error on line number %d, the marker %s does not exist.\n", fileName, j, l3);
		perr(bouf);
	      }
	      else {
		sprintf(bouf, "%s : Syntax error on line number %d, the marker %s does not exist.\n", fileName, j, l);
		perr(bouf);
	      }
    
	      // plus de nettoyage.

	      while (cons)
		{
		  con = cons->Next;
		  delete cons;
		  cons= con;
		}
	      delete lineres;
	      return vide;
	    }

	    con = cons;
	    cons = new Constraint(m1,m2,m3,pen);
	    cons->Next = con;
	  }	
      }
  
  fclose(fdata);

  // Allocation mmoire pour le tableau provisoire des champs de bit
  jeufus = new BioJeuPtr[numjeu + 1];
  jeufus[0] = NULL;

  // chargement diffrent...
  if ( Cross != Con ) {
    // Allocation mmoire pour le tableau des indices de marqueurs
    indmarq = new int[NbMarqueur + nbnm + 1];
    indmarq[0] = 0;

    // Allocation mmoire pour le tableau des merged
    mergemarq = new char[NbMarqueur + nbnm + 1];
    mergemarq[0] = 0;
    
    // Allocation mmoire pour le tableau des represents
    repmarq = new int[NbMarqueur + nbnm + 1];
    repmarq[0] = 0;

    // Allocation mmoire pour le tableau provisoire des noms de marqueurs
    nomarqfus = new charPtr[NbMarqueur + nbnm + 1];
    nomarq[0] = NULL;
      
    // Allocation mmoire pour le tableau provisoire des champs de bit
    bitjeufus = new int[NbMarqueur + nbnm + 1];
    bitjeufus[0] = 0;
      
    // remplissage
    for (i = 1; i <= NbMarqueur; i++)  {
      indmarq[i] = marqex[i];
      mergemarq[i]= Merged[i];
      repmarq[i] = Represents[i];
      nomarqfus[i] = NomMarq[i];

      // On copie masque du marqueur
      bitjeufus[i] = BitJeuMarq[i];
      // et s'il est en plus defini dans le nouveau jeu, on ajoute cette info.
      if (marqex[i]) bitjeufus[i] |= bitjeu;
    }

    for (i = 1; i <= nbnm; i++)  {
      indmarq[NbMarqueur + i] =  marqex[NbMarqueur + i];
      mergemarq[NbMarqueur + i] = 0;
      repmarq[NbMarqueur + i] = 0;
      nomarqfus[NbMarqueur + i] = nomarq[i];
      // le masque bit a bit est intiialise pour indiquer que ce
      // marqueur est defini dans le nouveu jeu
      bitjeufus[NbMarqueur + i] = bitjeu;
    }
      
    if ( NomMarq != NULL ) delete NomMarq;
    NomMarq = nomarqfus;
    delete nomarq;

    for (k = 1; k < nbm; k++) 
      delete nomarqj[k];
    delete nomarqj;

   if (Merged != NULL) delete Merged;
    Merged = mergemarq;

    if (Represents != NULL) delete Represents;
    Represents = repmarq;

    if ( BitJeuMarq != NULL ) delete BitJeuMarq;
    BitJeuMarq = bitjeufus;

    NbMarqueur += nbnm;
  } else  {
    nbm = 0;
    for (i = 1; i <= NbMarqueur; i++)
      if (marqex[i]) { 
	nbm++;
	BitJeuMarq[i] |= bitjeu;
      }
  }


  switch (Cross) 
    {
    case BC: 
      bjp = new BJS_BC(this,
		       numjeu,
                       fileName,
		       Cross, 
		       nbm, 
		       nbi, 
		       bitjeu,
		       indmarq,
		       Ech);
      
      sprintf(lineres, "%d f2 backcross %d %d %s", numjeu, nbm, nbi, fileName);
      break;

    case RISelf: 
      bjp = new BJS_BC(this,
		       numjeu,
                       fileName,
		       Cross, 
		       nbm, 
		       nbi, 
		       bitjeu,
		       indmarq,
		       Ech);
      bjp->Cross = RISelf;
      sprintf(lineres, "%d ri self %d %d %s", numjeu, nbm, nbi, fileName);
      break;
      
    case RISib: 
      bjp = new BJS_BC(this,
		       numjeu,
                       fileName,
		       Cross, 
		       nbm, 
		       nbi, 
		       bitjeu,
		       indmarq,
		       Ech);
      bjp->Cross = RISib;
      sprintf(lineres, "%d ri sib %d %d %s", numjeu, nbm, nbi, fileName);
      break;

    case IC: 
      bjp = new BJS_IC(this,
		       numjeu,
                       fileName,
		       Cross, 
		       nbm, 
		       nbi, 
		       bitjeu,
		       indmarq,
		       Ech);
      sprintf(lineres, "%d intercross %d %d %s", numjeu, nbm, nbi, fileName);
      break;

    case RH: 
      bjp = new BJS_RH(this,
		       numjeu,
                       fileName,
		       Cross, 
		       nbm, 
		       nbi, 
		       bitjeu,
		       indmarq,
		       Ech);
      sprintf(lineres, "%d haploid RH %d %d %s", numjeu, nbm, nbi, fileName);
      break;

    case RHD: 
      bjp = new BJS_RHD(this,
			numjeu,
                        fileName,
			Cross, 
			nbm, 
			nbi, 
			bitjeu,
			indmarq,
			Ech);
      sprintf(lineres, "%d diploid RH %d %d %s", numjeu, nbm, nbi, fileName);
      break;

    case Con: 
      bjp = new BJS_CO(this,
		       numjeu,
                       fileName,
		       Cross, 
		       nbm,
		       j-2,
		       bitjeu,
		       cons);
      sprintf(lineres, "%d constraint %d %s", numjeu, j-2, fileName);
      break;

    default :
      sprintf(bouf,  "Cartagene.cc::CharJeu() - pedigree impossible\n");
      pout(bouf);
      // exit(1);
      return vide;
      break;
    }

  ArbreJeu = bjp;

  for (i = 1; i <= NbJeu; i++) jeufus[i] = Jeu[i];
  jeufus[numjeu] = bjp;
  
  if ( Jeu != NULL ) delete Jeu;
  Jeu = jeufus;

  // chargement diffrent...
  if ( Cross != Con )
    {      
      // mise  jour de la liste des marqueurs actifs
      
      if ( MarkSelect != NULL ) delete MarkSelect;
      NbMS = nbm;
      MarkSelect = new int[nbm];
      j = 0;
      for (i = 1; i <= NbMarqueur; i++) 
	if (bitjeu & BitJeuMarq[i]) MarkSelect[j++] = i;
      
      // Le tas n'a plus de sens.

      int th = Heap.MaxHeapSize; 
      Heap.Init(this, th);
      
    }

  NbJeu = numjeu; 
  NbIndividu += nbi;

  delete marqex;
  delete vide;

  return lineres;
}

//-----------------------------------------------------------------------------
// Affichage de tous les marqueurs
//-----------------------------------------------------------------------------
// Paramtres : 
//-----------------------------------------------------------------------------

void CartaGene::AffMarq()
{
  int i, j;

  sprintf(bouf,  "\nMarkers :\n");
  pout(bouf);
  sprintf(bouf,  "--------:\n");
  pout(bouf);
  sprintf(bouf, "%3s %20s : %s", "Num", "Names", "Sets ");
  pout(bouf);

  for (j = 0; j < NbJeu-1; j++) {
    sprintf(bouf,  "    ");
    pout(bouf);
  }
  sprintf(bouf,  "Merges\n");
  pout(bouf);

  for (i = 1; i <= NbMarqueur; i++)  
    {
      sprintf(bouf, "%3d %20s :", i, NomMarq[i]);
      pout(bouf);
      for (j = 0; j < NbJeu; j++) 
	{
	  if ( BitJeuMarq[i] & 1<<j ) {
	    sprintf(bouf, " %3d", j + 1);
	    pout(bouf);
	  }
	  else {
	    sprintf(bouf, " %3s","");
	    pout(bouf);
	  }
	}
      sprintf(bouf,  Merged[i] ? "  merged " : "");
      pout(bouf);
      j = Represents[i];
      while (j != 0) {
	  sprintf(bouf, "%3d ",j);
	  pout(bouf);
	  j = Represents[j];
      }
      sprintf(bouf,  "\n");            
      pout(bouf);
    }  
}

//-----------------------------------------------------------------------------
// Affichage de tous les Jeux
//-----------------------------------------------------------------------------
// Paramtres : 
//-----------------------------------------------------------------------------

void CartaGene::AffJeu()
{

  BioJeu *bjp;

  sprintf(bouf,"\nData Sets :\n");
  pout(bouf);

  sprintf(bouf,"----------:\n");
  pout(bouf);
  
  sprintf(bouf,"%2s %16s %10s %11s %20s %11s %7s\n", 
	  "ID", 
	  "Data Type", 
	  "markers", 
	  "individuals",
	  "filename",
	  "constraints",
	  "merging");
  pout(bouf);

  for (int i = 1; i <= NbJeu; i++)  
  {
      sprintf(bouf,"%2d ", i);
      pout(bouf);
    
      switch (Jeu[i]->Cross) 
	{
	case BC: 
	  sprintf(bouf,"%16s ", "f2 backcross");
	  pout(bouf);
	  sprintf(bouf,"%10d %11d %20s\n",  
		  Jeu[i]->NbMarqueur, 
		  Jeu[i]->TailleEchant, 
		  Cart_basename(((BioJeuSingle *)Jeu[i])->NomJeu));
	  pout(bouf);
	  break;
	case RISelf:
	  sprintf(bouf,"%16s ", "ri self");
	  pout(bouf);
	  sprintf(bouf,"%10d %11d %20s\n", 
		  Jeu[i]->NbMarqueur, 
		  Jeu[i]->TailleEchant, 
		  Cart_basename(((BioJeuSingle *)Jeu[i])->NomJeu ));
	  pout(bouf);
	  break;
	case RISib:
	  sprintf(bouf,"%16s ", "ri sib");
	  pout(bouf);
	  sprintf(bouf,"%10d %11d %20s\n", Jeu[i]->NbMarqueur, Jeu[i]->TailleEchant, Cart_basename(((BioJeuSingle *)Jeu[i])->NomJeu));
	  pout(bouf);
	  break;
	case IC: 
	  sprintf(bouf,"%16s ", "intercross");
	  pout(bouf);
	  sprintf(bouf,"%10d %11d %20s\n", 
		  Jeu[i]->NbMarqueur, 
		  Jeu[i]->TailleEchant, 
		  Cart_basename(((BioJeuSingle *)Jeu[i])->NomJeu));
	  pout(bouf);
	  break;
	case RH: 
	  sprintf(bouf,"%16s ", "haploid RH");
	  pout(bouf);
	  sprintf(bouf,"%10d %11d %20s\n", 
		  Jeu[i]->NbMarqueur, 
		  Jeu[i]->TailleEchant, 
		  Cart_basename(((BioJeuSingle *)Jeu[i])->NomJeu));
	  pout(bouf);
	  break;
	case RHD: 
	  sprintf(bouf,"%16s ", "diploid RH");
	  pout(bouf);
	  sprintf(bouf,"%10d %11d %20s\n", 
		  Jeu[i]->NbMarqueur, 
		  Jeu[i]->TailleEchant, 
		  Cart_basename(((BioJeuSingle *)Jeu[i])->NomJeu));
	  pout(bouf);
	  break;
	case Mge: 
	  sprintf(bouf,"%16s ", "merged genetic");
	  pout(bouf);
	  sprintf(bouf,"%10d %11d %11s %20s %3d %3d\n", 
		  Jeu[i]->NbMarqueur, 
		  Jeu[i]->TailleEchant,
		  "",
		  "",
		  ((BJM_GE*)Jeu[i])->BJgauche->Id,
		  ((BJM_GE*)Jeu[i])->BJdroite->Id);
	  pout(bouf);
	  break;
	case Mor: 
	  sprintf(bouf,"%16s ", "merged by order");
	  pout(bouf);
	  sprintf(bouf,"%10d %11d %11s %20s %3d %3d\n", 
		  Jeu[i]->NbMarqueur, 
		  Jeu[i]->TailleEchant,
		  "",
		  "",
		  ((BJM_OR*)Jeu[i])->BJgauche->Id,
		  ((BJM_OR*)Jeu[i])->BJdroite->Id);
	  pout(bouf);
	  break;
	case Con: 
	  sprintf(bouf,"%16s ", "constraint");
	  pout(bouf);
	  sprintf(bouf,"%10d %11s %20s %11d\n", 
		  Jeu[i]->NbMarqueur,
		  "",
		  Cart_basename(((BioJeuSingle *)Jeu[i])->NomJeu),
		  Jeu[i]->TailleEchant);
	  pout(bouf);
	  break;
	default:
	  break;
	}
    }  
}

//-----------------------------------------------------------------------------
// retourne la liste des chaines d'info sur les jeux de donnes
//-----------------------------------------------------------------------------
// Paramtres : 
//-----------------------------------------------------------------------------

char ** CartaGene::GetArbre()
{

  char* lineres;

  // allocation

  char** arbre =  new char*[NbJeu + 1];
  arbre[NbJeu] = NULL;
  
  for (int i = 1; i <= NbJeu; i++)  
    {
      
      lineres = new char[1024];
      
      switch (Jeu[i]->Cross) 
	{
	case BC: 
	  sprintf(lineres, "%d \"f2 backcross\" %s %d %d", 
		  i, 
		  ((BioJeuSingle *)Jeu[i])->NomJeu,
		  Jeu[i]->NbMarqueur, 
		  Jeu[i]->TailleEchant);
	  break;
	  
	case RISelf: 
	  
	  sprintf(lineres, "%d \"ri self\" %s %d %d", 
		  i,
		  ((BioJeuSingle *)Jeu[i])->NomJeu,
		  Jeu[i]->NbMarqueur, 
		  Jeu[i]->TailleEchant);
	  break;
	  
	case RISib: 
	  
	  sprintf(lineres, "%d \"ri sib\" %s %d %d", 
		  i, 
		  ((BioJeuSingle *)Jeu[i])->NomJeu,
		  Jeu[i]->NbMarqueur, 
		  Jeu[i]->TailleEchant);
	  
	case IC: 
	  
	  sprintf(lineres, "%d \"intercross\" %s %d %d", 
		  i, 
		  ((BioJeuSingle *)Jeu[i])->NomJeu,
		  Jeu[i]->NbMarqueur, 
		  Jeu[i]->TailleEchant);
	  break;
	case RH: 
	  
	  sprintf(lineres, "%d \"radiated hybrid\" %s %d %d", 
		  i, 
		  ((BioJeuSingle *)Jeu[i])->NomJeu,
		  Jeu[i]->NbMarqueur, 
		  Jeu[i]->TailleEchant);
	  break;
	case RHD: 
	  
	  sprintf(lineres, "%d \"radiated hybrid diploid\" %s %d %d", 
		  i, 
		  ((BioJeuSingle *)Jeu[i])->NomJeu,
		  Jeu[i]->NbMarqueur, 
		  Jeu[i]->TailleEchant);
	  break;
	case Con: 
	  
	  sprintf(lineres, "%d \"constraint\" %s %d", 
		  i, 
		  ((BioJeuSingle *)Jeu[i])->NomJeu,
		  Jeu[i]->TailleEchant);
	  break;
	case Mge: 

	  sprintf(lineres, "%d \"merged genetic\" %d %d %d %d", 
		  i, 
		  ((BioJeuMerged *)Jeu[i])->BJgauche->Id,
		  ((BioJeuMerged *)Jeu[i])->BJdroite->Id,
		  Jeu[i]->NbMarqueur, 
		  Jeu[i]->TailleEchant);
	  break;
	case Mor: 
	  
	  sprintf(lineres, "%d \"merged by order\" %d %d %d %d", 
		  i, 
		  ((BioJeuMerged *)Jeu[i])->BJgauche->Id,
		  ((BioJeuMerged *)Jeu[i])->BJdroite->Id,
		  Jeu[i]->NbMarqueur, 
		  Jeu[i]->TailleEchant);
	  break;
	default :
	  
	  break;
	}
      
      arbre[i-1] = lineres;     
    }

  return (arbre);
}

//-----------------------------------------------------------------------------
// Affichage de l'chantillon d'un jeu
//-----------------------------------------------------------------------------
// Paramtres :
// - le numro du jeu 
//-----------------------------------------------------------------------------

void CartaGene::DumpEch(int nbjeu)
{
  sprintf(bouf, "\nDump Data Set %2d :\n", nbjeu);
  pout(bouf);
  sprintf(bouf,  "-----------------:\n");
  pout(bouf);
  Jeu[nbjeu]->DumpEch();
}

//-----------------------------------------------------------------------------
// Affichage des LOD 2pt d'un jeu
//-----------------------------------------------------------------------------
// Paramtres : 
// - le numro du jeu 
//-----------------------------------------------------------------------------

void CartaGene::DumpTwoPointsLOD(int nbjeu)
{
  sprintf(bouf, "\nDump two points LOD %2d :\n", nbjeu);
  pout(bouf);
  sprintf(bouf,  "-----------------------:\n");
  pout(bouf);
  Jeu[nbjeu]->DumpTwoPointsLOD();
}

//-----------------------------------------------------------------------------
// Affichage des LOD 2pt de la slection
//-----------------------------------------------------------------------------
// Paramtres : 
//-----------------------------------------------------------------------------

void CartaGene::PrintTwoPointsLOD()
{
  int i, j;

  // vrification de base
  
  if (NbMS == 0) 
    {
      sprintf(bouf,  "Error : Empty selection of loci.\n");
      perr(bouf);
      return;
    }
  
  sprintf(bouf,  "\n          ");
  pout(bouf);
  for (i = 0; i < NbMS; i++) {
    sprintf(bouf, "%5d ",MarkSelect[i]);
    pout(bouf);
  }
  fflush(stdout);
  sprintf(bouf,  "\n          ");
  pout(bouf);
  
  for (i = 0; i < NbMS; i++) {
    sprintf(bouf, "%5.5s ",NomMarq[MarkSelect[i]]);
    pout(bouf);
  }
  sprintf(bouf,  "\n");
  pout(bouf);
  fflush(stdout);

  sprintf(bouf,  "          ");
  pout(bouf);
  for (i = 0; i < NbMS; i++) {
    sprintf(bouf,  "------");
    pout(bouf);
  }
  sprintf(bouf,  "\n");
  pout(bouf);
  fflush(stdout);
  for (i = 0; i< NbMS; i++) {      
    sprintf(bouf, "%8.8s |",NomMarq[MarkSelect[i]]);
    pout(bouf);
    fflush(stdout);
    for (j = 0; j < NbMS; j++)
      if (i == j)  {
	sprintf(bouf,  "------");
	pout(bouf);
      }
      else {
	sprintf(bouf, "%5.1f ", GetTwoPointsLOD(MarkSelect[i], MarkSelect[j]));
	pout(bouf);
      }
    sprintf(bouf,  "\n");
    pout(bouf);
    fflush(stdout);
  }
}

//-----------------------------------------------------------------------------
// Affichage des LOD 2pt de la slection
//-----------------------------------------------------------------------------
// Paramtres : 
//-----------------------------------------------------------------------------
void CartaGene::DumpDouble()
{
  int i, j,count = 0;

  sprintf(bouf,  "\nPossible double markers:");
  pout(bouf);
  for (i = 0; i< NbMS-1; i++) 
    for (j = i; j < NbMS; j++)
      if ((i != j) && 
	  ArbreJeu->Compatible(MarkSelect[i], MarkSelect[j]) &&
	  GetTwoPointsLOD(MarkSelect[i], MarkSelect[j])> 0.0) {
	if (!count) {
	  sprintf(bouf,  "\n\n");
	  pout(bouf);
	}
	sprintf(bouf, "    %15s = %-15s [%.1f]\n",NomMarq[MarkSelect[i]], 
	       NomMarq[MarkSelect[j]],GetTwoPointsLOD(MarkSelect[i], MarkSelect[j]));
	pout(bouf);
	count++;
      }
  if (!count) {
    sprintf(bouf,  " none.");
    pout(bouf);
  }
  sprintf(bouf,  "\n");
  pout(bouf);
  fflush(stdout);
}
//-----------------------------------------------------------------------------
// Affichage des Dist 2pt de la slection
//-----------------------------------------------------------------------------
// Paramtres : 
//-----------------------------------------------------------------------------

void CartaGene::PrintTwoPointsDist(char unit[2])
{
  
  if (NbMS == 0) 
    {
      sprintf(bouf,  "Error : Empty selection of loci.\n");
      perr(bouf);
      return;
    }
  
  switch(unit[0]) 
    {
    case 'h': break;
    case 'k': break;
    default: 
      sprintf(bouf,  "Error : UnitFlag should be 'k' or 'h'.\n");
      perr(bouf);
      return;
    }
  
  sprintf(bouf,  "\nPrint two points distance matrices of the loci selection :\n");
  pout(bouf);
  sprintf(bouf,  "---------------------------------------------------------:\n");
  pout(bouf);
  fflush(stdout);

  ArbreJeu->PrintTwoPointsDist(unit);
  
  fflush(stdout);
}

//-----------------------------------------------------------------------------
// Affichage des FR 2pt de la slection
//-----------------------------------------------------------------------------
// Paramtres : 
//-----------------------------------------------------------------------------

void CartaGene::PrintTwoPointsFR()
{
  int i, j;

  // vrification de base
  
  if (NbMS == 0) 
    {
      sprintf(bouf,  "Error : Empty selection of loci.\n");
      perr(bouf);
      return;
    }
  
  sprintf(bouf,  "\nPrint two points recombination fractions  matrices of the loci selection :\n");
  pout(bouf);
  sprintf(bouf,  "---------------------------------------------------------------------------:\n");
  pout(bouf);
  fflush(stdout);
  sprintf(bouf,  "\n       ");
  pout(bouf);
  
  for (i = 0; i < NbMS; i++) {
    sprintf(bouf, "%5.5s ",NomMarq[MarkSelect[i]]);
    pout(bouf);
  }
  sprintf(bouf,  "\n");
  pout(bouf);
  fflush(stdout);

  sprintf(bouf,  "       ");
  pout(bouf);
  for (i = 0; i < NbMS; i++)  {
    sprintf(bouf,  "------");
    pout(bouf);
  }
  sprintf(bouf,  "\n");
  pout(bouf);
  fflush(stdout);
  for (i = 0; i< NbMS; i++) {      
    sprintf(bouf, "%5.5s |",NomMarq[MarkSelect[i]]);
    pout(bouf);
    fflush(stdout);
    for (j = 0; j < NbMS; j++)
      if (i == j)  {
	sprintf(bouf,  "------");
	pout(bouf);
      }
      else {
	sprintf(bouf, "%5.1f ", GetTwoPointsFR(MarkSelect[i], MarkSelect[j]));
	pout(bouf);
      }
    sprintf(bouf,  "\n");
    pout(bouf);
    fflush(stdout);
  }
  fflush(stdout);
}


//-----------------------------------------------------------------------------
// Affichage des FR d'un jeu
//-----------------------------------------------------------------------------
// Paramtres : 
// - le numro du jeu 
//-----------------------------------------------------------------------------

void CartaGene::DumpTwoPointsFR()
{
  sprintf(bouf,  "Dump two points :\n");
  pout(bouf);
  
  ArbreJeu->DumpTwoPointsFR();

  /*  Jeu[nbjeu]->DumpTwoPointsFR();*/
}

//-----------------------------------------------------------------------------
// Calcul et Affichage des groupes d'un jeu
//-----------------------------------------------------------------------------
// Paramtres : 
// - disthres seuil de la distance
// - lodthres seuil du lod
// Valeur de retour : le nombre de goupes trouvs
//-----------------------------------------------------------------------------

int CartaGene::Groupe(double disthres, double lodthres)
{

  if (ArbreJeu == NULL) 
    {
      sprintf(bouf,  "Error : No Data Set Loaded..\n");
      perr(bouf);
      return -1;
    }

  return ArbreJeu->Groupe(disthres, lodthres);
}


//-----------------------------------------------------------------------------
// Affichage de tous l'chantillon
//-----------------------------------------------------------------------------
// Paramtres : 
// - Les indices des deux jeux  fusionner
// -// Valeur de retour : Chaine de car. contenant 
// - le numro du jeu
// - le type du jeu
// - le nombre de marqeurs
// - le nombre d'individus
// ou une chaine vide en cas d'erreur.
//-----------------------------------------------------------------------------

char *CartaGene::MerGen(int gauche, int droite) 
{

  char *lineres ;
  int i,j;

  CrossType Cross;
  int bitjeu;
  int nbm = 0; 
  BioJeu **jeufus;
  char *vide;

  // pour retourner une liste vide(voir conventions)
  vide = new char[1];
  vide[0] = '\0';

  // test de l'existence des jeux  fusionner

  if (gauche > NbJeu || gauche < 1) 
    {
      sprintf(bouf, "Error : Unknown Data Set %d.\n", gauche);
      perr(bouf);
      return vide;
    }

  if (droite > NbJeu || droite < 1) 
    {
      sprintf(bouf, "Error : Unknown Data Set %d.\n", droite);
      perr(bouf);
      return vide;
    }

  // interdiction de merger gntiquement des mergs sur l'ordre. 

  if (Jeu[gauche]->Cross == Mor || Jeu[gauche]->Cross == Con)
    {
      sprintf(bouf, "Error on Data Set %d : The type is not allowed.\n", gauche);
      perr(bouf);
      return vide;
    }

  if (Jeu[droite]->Cross == Mor || Jeu[droite]->Cross == Con)
    {
      sprintf(bouf, "Error on Data Set %d : The type merged is not allowed.\n", droite);
      perr(bouf);
      return vide;
    }

  // RH ou Pas

  if (Jeu[gauche]->HasRH() != Jeu[droite]->HasRH())
    {
      sprintf(bouf,  "Error : Merging radiated hybrid with another type is not allowed.\n");
      perr(bouf);
      return vide;
    }

  if (Jeu[droite]->HasRH() && (Jeu[droite]->Cross != Jeu[gauche]->Cross))
    {
      sprintf(bouf, "Error on Data Set %d : Cannot merge diploid and haploid data.\n", droite);
      perr(bouf);
      return vide;
    }

  //RI ou pas
  
  if (Jeu[gauche]->HasRI() != Jeu[droite]->HasRI())
    {
      sprintf(bouf,  "Error : Merging RISib or RISelf with another type is not allowed.\n");
      perr(bouf);
      return vide;
    }
  
  if (Jeu[gauche]->HasRI() && (Jeu[droite]->Cross != Jeu[gauche]->Cross))
    {
      sprintf(bouf,  "Error : Merging RISib with RISelf is not allowed.\n");
      perr(bouf);
      return vide;
    }
  
  lineres = new char[256];
  NbJeu++;
  Cross = Mge;
  bitjeu = 1<<(NbJeu - 1);

  for (i = 1; i <= NbMarqueur; i++) 
    if ((BitJeuMarq[i] & Jeu[gauche]->BitJeu) || 
	(BitJeuMarq[i] & Jeu[droite]->BitJeu))
      {
	BitJeuMarq[i] |= bitjeu;
	nbm++;
      }
  
  ArbreJeu = new BJM_GE(this,
			NbJeu,
 			Cross,
 			nbm,
 			bitjeu,
 			Jeu[gauche], 
 			Jeu[droite]);

  // Allocation mmoire pour le tableau provisoire des champs de bit
  jeufus = new BioJeuPtr[NbJeu + 1];
  jeufus[0] = NULL;
  
  for (i = 1; i < NbJeu; i++) jeufus[i] = Jeu[i];
  jeufus[NbJeu] = ArbreJeu;
  
  delete Jeu;
  Jeu = jeufus;

  // mise  jour de la liste des marqueurs actifs
  
  delete MarkSelect;
  NbMS = nbm;
  MarkSelect = new int[nbm];
  j = 0;
  for (i = 1; i <= NbMarqueur; i++) 
    if ((BitJeuMarq[i] & Jeu[gauche]->BitJeu) || 
	(BitJeuMarq[i] & Jeu[droite]->BitJeu))
      MarkSelect[j++] = i;

  sprintf(lineres, "%d merged genetic %d %d", NbJeu, nbm, 
	  Jeu[gauche]->TailleEchant + Jeu[droite]->TailleEchant);

  delete vide;

  // Le tas n'a plus de sens.

  int th = Heap.MaxHeapSize; 
  Heap.Init(this, th);

  return lineres;

}
//-----------------------------------------------------------------------------
// Affichage de tous l'chantillon
//-----------------------------------------------------------------------------
// Paramtres : 
// - Les indices des deux jeux  fusionner
// -// Valeur de retour : Chaine de car. contenant 
// - le numro du jeu
// - le type du jeu
// - le nombre de marqeurs
// - le nombre d'individus
// ou une chaine vide en cas d'erreur.
//-----------------------------------------------------------------------------

char *CartaGene::MergOr(int gauche, int droite) 
{

  char *lineres ;
  int i,j;

  CrossType Cross;
  int bitjeu;
  int nbm = 0; 
  BioJeu **jeufus;
  char *vide;

  // pour retourner une liste vide(voir conventions)
  vide = new char[1];
  vide[0] = '\0';
  
  // test de l'existence des jeux  fusionner
  
  if (gauche > NbJeu || gauche < 1) 
    {
      sprintf(bouf, "Error : Unknown Data Set %d.\n", gauche);
      perr(bouf);
      return vide;
    }

  if (droite > NbJeu || droite < 1) 
    {
      sprintf(bouf, "Error : Unknown Data Set %d.\n", droite);
      perr(bouf);
      return vide;
    }

  lineres = new char[256];
  NbJeu++;
  Cross = Mor;
  bitjeu = 1<<(NbJeu - 1);

  for (i = 1; i <= NbMarqueur; i++) 
    if ((BitJeuMarq[i] & Jeu[gauche]->BitJeu && Jeu[gauche]->Cross != Con) || 
	(BitJeuMarq[i] & Jeu[droite]->BitJeu && Jeu[droite]->Cross != Con))
      {
	BitJeuMarq[i] |= bitjeu;
	nbm++;
      }

  ArbreJeu = new BJM_OR(this,
			NbJeu,
 			Cross,
 			nbm,
 			bitjeu,
 			Jeu[gauche], 
 			Jeu[droite]);

  // Allocation mmoire pour le tableau provisoire des champs de bit
  jeufus = new BioJeuPtr[NbJeu + 1];
  jeufus[0] = NULL;
  
  for (i = 1; i < NbJeu; i++) jeufus[i] = Jeu[i];
  jeufus[NbJeu] = ArbreJeu;
  
  delete Jeu;
  Jeu = jeufus;

  // mise  jour de la liste des marqueurs actifs
  
  delete MarkSelect;
  NbMS = nbm;
  MarkSelect = new int[nbm];
  j = 0;
  for (i = 1; i <= NbMarqueur; i++) 
    if (BitJeuMarq[i] & bitjeu)
      MarkSelect[j++] = i;
  
  sprintf(lineres, "%d merged by order %d %d", NbJeu, nbm, 
	  Jeu[gauche]->TailleEchant + Jeu[droite]->TailleEchant);

  delete vide;

  // Le tas n'a plus de sens.

  int th = Heap.MaxHeapSize; 
  Heap.Init(this, th);
  
  return lineres;

}

//-----------------------------------------------------------------------------
// Accs  une valeur LOD score 2pt.
//-----------------------------------------------------------------------------
// Paramtres : 
//-----------------------------------------------------------------------------
double CartaGene::GetTwoPointsLOD(int m1, int m2) const
{
  return ArbreJeu->GetTwoPointsLOD(m1, m2);
}

//-----------------------------------------------------------------------------
// Accs  la fraction de recombinaison 2pt.
//-----------------------------------------------------------------------------
// Paramtres : 
//-----------------------------------------------------------------------------
double CartaGene::GetTwoPointsFR(int m1, int m2) const
{
  return ArbreJeu->GetTwoPointsFR(m1, m2);
}

//-----------------------------------------------------------------------------
// Accs  la distance 2pt.
//-----------------------------------------------------------------------------
// Paramtres : 
//-----------------------------------------------------------------------------
double CartaGene::GetTwoPointsDH(int m1, int m2) const
{
  return ArbreJeu->GetTwoPointsDH(m1, m2);
}

//-----------------------------------------------------------------------------
//  Calcul de la vraisemblance maximum par EM.
//-----------------------------------------------------------------------------
// Paramtres : 
// - une carte
// - seuil de convergence
// Valeur de retour : la vraisemblance maximum
//-----------------------------------------------------------------------------

double CartaGene::ComputeEM(Carte *map)
{
  return ArbreJeu->ComputeEM(map);
}

//-----------------------------------------------------------------------------
//  Calcul de la vraisemblance maximum par EM.
//  Une interface plus sophistiquee:
//  On converge jusqu'a epsilon1. Si la vraisemblance est inferieure a
//  threshold on laisse tomber, sinon on converge a epsilon2
//-----------------------------------------------------------------------------
// Paramtres : 
// - une carte
// - niveau a partir duquel on convergera a epsilon2
// - premier seuil de convergence (epsilon1)
// - second seuil de convergence (epsilon2)
// Valeur de retour : la vraisemblance maximum.
//-----------------------------------------------------------------------------

double CartaGene::ComputeEMS(Carte *map, 
			     double threshold)
{

  return ArbreJeu->ComputeEMS(map, threshold);
  
}

//-----------------------------------------------------------------------------
//  Affichage simple slectif d'une carte
// Seuls les marqueurs du gnotype sont affichs
//-----------------------------------------------------------------------------
// Paramtres : 
// - une carte
// Valeur de retour : 
//-----------------------------------------------------------------------------

void CartaGene::PrintMap(Carte *map) const
{
  sprintf(bouf, "\nMap %2d : log10-likelihood = %8.2f\n", 
	 map->Id, 
	 map->coutEM);
  pout(bouf);
  sprintf(bouf,  "-------:\n");
  pout(bouf);
  sprintf(bouf, "%4s : Marker List ...\n", "Set");
  pout(bouf);
  ArbreJeu->PrintMap(map);
}

//-----------------------------------------------------------------------------
//  Affichage dtaill d'une carte
// Seuls les marqueurs du gnotype sont affichs
//-----------------------------------------------------------------------------
// Paramtres : 
// - une carte
// Valeur de retour : 
//-----------------------------------------------------------------------------

void CartaGene::PrintDMap(Carte *map, int envers, Carte *mapref) const
{
  sprintf(bouf, "\nMap %2d : log10-likelihood = %8.2f, log-e-likelihood = %8.2f\n", 
	 map->Id, map->coutEM, map->coutEM * log(10));
  pout(bouf);
  sprintf(bouf,  "-------:\n");
  pout(bouf);
  ArbreJeu->PrintDMap(map, envers, mapref);
}

//-----------------------------------------------------------------------------
// Affichage dtaill des deltas carteFW / autres marqueurs
//-----------------------------------------------------------------------------
// Paramtres : 
// - La structure des deltas
// - la carte FW
// - Le nombre de marqueur de la carte FW
// - Le nombre de marqueur a testee
// Valeur de retour :
//----------------------------------------------------------------------------
  
void CartaGene::PrintVMap(struct testmartin *pstruct, Carte *data, double *distd, int nbm, int nbtas, double AM)
{
  // nbm = nb de marqueur dans le framework
  // nbtas = nb de marqueur a tester
  int i,j,aprescomp,avantcomp;
  int avantbest,apresbest; // avantbest apresbest   num marqueur avant,apres la meilleur position
  unsigned int maxl=0;
  char form[128],form1[128],*formb;
  double Lodapres,Lodavant,Dhapres,Dhavant,Dhbest,Dhprop;

  formb=" ";

  for (i = 0; i < nbtas; i++)
    {
      j = Represents[pstruct[i].numero];
      while (j != 0)
      {
        if (strlen(NomMarq[j]) > maxl)
          maxl = strlen(NomMarq[j]);
        j = Represents[j];
      }
      if (strlen(NomMarq[pstruct[i].numero]) > maxl)
	maxl = strlen(NomMarq[pstruct[i].numero]);
    }

  maxl = (maxl < 5 ? 5 : maxl);

  sprintf(form, " %c%d%c |",'%', maxl, 's');

  // 1ere ligne
  sprintf(bouf, form,formb);
  pout(bouf);
  for (i = 0; i < nbm; i++) 
    if ((data->ordre[i] - (data->ordre[i]%100))/100) {
      sprintf(bouf, " %d", (data->ordre[i] - (data->ordre[i]%100))/100);
      pout(bouf);
    }
    else {
      sprintf(bouf,  "  "); 
      pout(bouf);
    }

  sprintf(bouf,  "  |\n");
  pout(bouf);
  // 2eme ligne
  sprintf(bouf, form,formb);
  pout(bouf);

  for (i = 0; i < nbm ; i++)
    if ((data->ordre[i]%100 - ((data->ordre[i]%100)%10))/10 || data->ordre[i] > 99) {
      sprintf(bouf, " %d", (data->ordre[i]%100 - ((data->ordre[i] %100)%10))/10);
      pout(bouf);
    }
    else {
      sprintf(bouf,  "  ");
      pout(bouf);
    }

  sprintf(bouf,  "  |     Lod2pt         Dist2pt\n");
  pout(bouf);
  // 3eme ligne
  sprintf(bouf, form,formb);
  pout(bouf);


  for (i = 0; i < nbm; i++) {  
    sprintf(bouf, " %d", data->ordre[i]%10);
    pout(bouf);
  }

  sprintf(bouf,  "  |  Left<-M->Right Left<-M->Right | 0->N    N->M | Weight Nb:W<AM  Id  | Name\n");
  pout(bouf);
  // 4eme ligne  les tirets
  sprintf(form, "%c%d%c--|",'%', maxl, 's');
  sprintf(bouf, form,formb);
  pout(bouf);

  for (i = 0; i < nbm; i++) {  
    sprintf(bouf,  "--");
    pout(bouf);
  }

  sprintf(bouf,  "--|--------------------------------|--------------|---------------------|-------\n");
  pout(bouf);

  for (i = 0; i < nbtas; i++)  
  {
    sprintf(form, " %c%d%c |",'%', maxl, 's');
    sprintf(bouf, form, NomMarq[pstruct[i].numero]);
    pout(bouf);

    for (j = 0; j < nbm+1 ; j++)
    {
        if ((BitJeuMarq[pstruct[i].numero] & ArbreJeu->BitJeu) && (pstruct[i].pdelta[j] <= AM))
        {
          if (j == pstruct[i].meilleur)
          {
  	    sprintf(bouf,  "+ ");
  	  }
          else
          {
	    sprintf(bouf, "%d ",pstruct[i].pdelta[j]);
	  }
        }
        else
        {
	  sprintf(bouf,  "  ");
        }
       pout(bouf);

      if (j == pstruct[i].meilleur)
      {
	avantbest=data->ordre[j-1];
        apresbest=data->ordre[j];
        Lodavant=0.0;Lodapres=0.0;
        Dhapres=0.0;
        Dhavant=0.0;
        Dhbest=0.0;
        Dhprop=0.0;
	// Bout Gauche
        if ( j == 0 )
        {
          avantbest=0;
          // marqueur apres pas compatible
          if (!((BitJeuMarq[apresbest] & ArbreJeu->BitJeu) && (BitJeuMarq[pstruct[i].numero] & ArbreJeu->BitJeu)))
          {
            apresbest=0;
	    sprintf(form, "|    -       -        -     -    | L End    NA< |  %4d  %4d     %-4d",pstruct[i].poids,pstruct[i].nbcas,pstruct[i].numero);
          }
          // marqueur apres compatible
          else
	  {
            Lodapres=GetTwoPointsLOD(apresbest,pstruct[i].numero);
            Dhapres=(100 * GetTwoPointsDH(apresbest,pstruct[i].numero));
            Dhprop=Dhapres;
            // si sur marqueur apres 0->N=Dhapres et N->M=0
            if ( Dhapres == 0)
	       sprintf(form, "|    -    %6.2f      -  %5.1f   | %5.1f   0    |  %4d  %4d     %-4d",Lodapres,Dhapres,Dhprop,pstruct[i].poids,pstruct[i].nbcas,pstruct[i].numero);
	    else
 	       sprintf(form, "|    -    %6.2f      -  %5.1f   |   -    %5.1f |  %4d  %4d     %-4d",Lodapres,Dhapres,Dhprop,pstruct[i].poids,pstruct[i].nbcas,pstruct[i].numero);
          }
        }
        // Pas Bout Gauche
 	else
          // Bout Droit
	  if ( j == nbm)
	  {
            apresbest=0;
            // marqueur avant pas compatible
	    if (!((BitJeuMarq[avantbest] & ArbreJeu->BitJeu) && (BitJeuMarq[pstruct[i].numero] & ArbreJeu->BitJeu)))
            {
              avantbest=0;
	      sprintf(form, "|    -       -        -     -    | R End    NA> |  %4d  %4d     %-4d",pstruct[i].poids,pstruct[i].nbcas,pstruct[i].numero);
            }
	    // marqueur avant compatible
            else
	    {
              Lodavant=GetTwoPointsLOD(avantbest,pstruct[i].numero);
              Dhavant=(100 * GetTwoPointsDH(avantbest,pstruct[i].numero));
              Dhbest=distd[j-1];
              Dhprop=Dhavant;
	      // si sur marqueur avant 0->N=Dhbest et N->M=0
              if ( Dhavant == 0)
	        sprintf(form, "| %6.2f     -     %5.1f    -    | %5.1f   0    |  %4d  %4d     %-4d",Lodavant,Dhavant,Dhbest,pstruct[i].poids,pstruct[i].nbcas,pstruct[i].numero);
       	      else
 	        sprintf(form, "| %6.2f     -     %5.1f    -    | %6.1f %5.1f |  %4d  %4d     %-4d",Lodavant,Dhavant,Dhbest,Dhprop,pstruct[i].poids,pstruct[i].nbcas,pstruct[i].numero);
            }
          }
	  // Ni Bout Gauche Ni Bout Droit
          else
          {
            // marqueur avant pas compatible
	    if (!((BitJeuMarq[avantbest] & ArbreJeu->BitJeu) && (BitJeuMarq[pstruct[i].numero] & ArbreJeu->BitJeu)))
            {
              avantbest=0;
	      // marqueur apres pas compatible
	      if (!((BitJeuMarq[apresbest] & ArbreJeu->BitJeu) && (BitJeuMarq[pstruct[i].numero] & ArbreJeu->BitJeu)))
              { 
                apresbest=0;
 	        sprintf(form, "|    -       -        -     -    |   -       -  |  %4d  %4d     %-4d",pstruct[i].poids,pstruct[i].nbcas,pstruct[i].numero);
              }
	      // marqueur apres compatible
              else
	      { 
                Lodapres=GetTwoPointsLOD(apresbest,pstruct[i].numero);
                Dhapres=(100 * GetTwoPointsDH(apresbest,pstruct[i].numero));
                Dhbest=distd[j];                                           // ditance cumulee du marqueur apres
                sprintf(form, "|   nc    %6.2f     nc  %5.1f   | %6.1f  NA<  |  %4d  %4d     %-4d",Lodapres,Dhapres,Dhbest,pstruct[i].poids,pstruct[i].nbcas,pstruct[i].numero);
              }
            }
	    // marqueur avant compatible
            else
	    { 
	      // marqueur apres pas compatible
	      if (!((BitJeuMarq[apresbest] & ArbreJeu->BitJeu) && (BitJeuMarq[pstruct[i].numero] & ArbreJeu->BitJeu)))
              { 
                Lodavant=GetTwoPointsLOD(avantbest,pstruct[i].numero);
                Dhavant=(100 * GetTwoPointsDH(avantbest,pstruct[i].numero));
                Dhbest=distd[j-1];                                         // ditance cumulee du marqueur avant
                sprintf(form, "| %6.2f    nc     %5.1f   nc    | %6.1f   NA> |  %4d  %4d     %-4d",Lodavant,Dhavant,Dhbest,pstruct[i].poids,pstruct[i].nbcas,pstruct[i].numero);
              }
	      // marqueur apres compatible
              else
	      { 
                Lodapres=GetTwoPointsLOD(apresbest,pstruct[i].numero);
                Dhapres=(100 * GetTwoPointsDH(apresbest,pstruct[i].numero));
                Lodavant=GetTwoPointsLOD(avantbest,pstruct[i].numero);
                Dhavant=(100 * GetTwoPointsDH(avantbest,pstruct[i].numero));
                Dhbest=distd[j-1];                                         // distance cumulee du marqueur avant
                Dhprop=((distd[j]-distd[j-1])*Dhavant)/(Dhavant+Dhapres);  // marqueur m  (dist(n,n+1)*dist(n,m)) / (dist(n,m) + dist(m,n+1))
		// si sur marqueur avant 0->N=Dhbest et N->M=0
                if ( Dhavant == 0)
		   sprintf(form, "| %6.2f  %6.2f   %5.1f %5.1f   | %6.1f  0    |  %4d  %4d     %-4d",Lodavant,Lodapres,Dhavant,Dhapres,Dhbest,pstruct[i].poids,pstruct[i].nbcas,pstruct[i].numero);
	        else
		   // si sur marqueur apres 0->N=distd[j] et N->M=0
		   if ( Dhapres == 0)
		      sprintf(form, "| %6.2f  %6.2f   %5.1f %5.1f   | %6.1f  0    |  %4d  %4d     %-4d",Lodavant,Lodapres,Dhavant,Dhapres,distd[j],pstruct[i].poids,pstruct[i].nbcas,pstruct[i].numero);
		   else
		      sprintf(form, "| %6.2f  %6.2f   %5.1f %5.1f   | %6.1f %5.1f |  %4d  %4d     %-4d",Lodavant,Lodapres,Dhavant,Dhapres,Dhbest,Dhprop,pstruct[i].poids,pstruct[i].nbcas,pstruct[i].numero);
              }
	    }
	  }
       } // Fin if j == pstruct
     } // Fin for j=

     sprintf(bouf, form);
     pout(bouf);
     sprintf(form, "| %c-%d%c\n",'%', maxl, 's');
     sprintf(bouf, form, NomMarq[pstruct[i].numero]);
     pout(bouf);
  }
}




//-----------------------------------------------------------------------------
//  Affichage simple slectif d'une carte
// Seuls les marqueurs du gnotype sont affichs
//-----------------------------------------------------------------------------
// Paramtres : 
// - une carte
// Valeur de retour : 
//-----------------------------------------------------------------------------

void CartaGene::PrintMap(int nbmap) const
{

  if (Heap.HeapSize == 0) 
    {
      sprintf(bouf,  "Error : Empty heap.\n");
      perr(bouf);
      return;
    }

  if ( nbmap < 0 || nbmap >= Heap.HeapSize )
    {
      sprintf(bouf,  "Error : This map does not exist.\n");
      perr(bouf);
      return;
    }
    
  sprintf(bouf, "\nMap %2d : log10-likelihood = %8.2f\n", 
	 (Heap.MapFromId(nbmap))->Id, 
	 (Heap.MapFromId(nbmap))->coutEM);
  pout(bouf);
  sprintf(bouf,  "-------:\n");
  pout(bouf);
  sprintf(bouf, "%4s : Marker List ...\n", "Set");
  pout(bouf);
  ArbreJeu->PrintMap(Heap.MapFromId(nbmap));
}

//-----------------------------------------------------------------------------
//  
//-----------------------------------------------------------------------------
// Paramtres : 
// - une carte
// Valeur de retour : une liste
//-----------------------------------------------------------------------------


char *** CartaGene::GetMap(char unit[2],
			   int nbmap) const
{
  
  // pour retourner une liste vide(voir conventions)
  //char * vide = new char[1];
  //vide[0] = '\0';

  switch(unit[0]) 
    {
    case 'h': break;
    case 'k': break;
    default: 
      sprintf(bouf,  "Error : UnitFlag should be 'k' or 'h'.\n");
      perr(bouf);
      return NULL;
    }

  if (Heap.HeapSize == 0) 
    {
      sprintf(bouf,  "Error : Empty heap.\n");
      perr(bouf);
      return NULL;
    }
  
  if ( nbmap < 0 || nbmap >= Heap.HeapSize )
    {
      sprintf(bouf,  "Error : This map does not exist.\n");
      perr(bouf);
      return NULL;
    }

  char*** lm = ArbreJeu->GetMap(unit, Heap.MapFromId(nbmap));
  int i = 0;

  // dcompte des cartes
  
  while (lm[i++] != NULL);

  // allocation

  char*** lmn =  new char**[i + 1];
  lmn[i] = NULL;

  char ** log = new char*[3];
  log[2] = NULL;
  // remplissage du log gnral
  
  char * temp = new char[8];
  sprintf(temp, "%.2f", 
	  (Heap.MapFromId(nbmap))->coutEM);
  log[1] = temp;

  // remplissage de l'indice de la carte
  
  temp = new char[8];
  sprintf(temp, "%d", nbmap);
  log[0] = temp;

  lmn[0] = log;

  i = 0; 
  // remplissage des cartes
  while (lm[i] != NULL)
    lmn[i+1] =  lm[i++];  

  delete [] lm;

  return(lmn);
    
}
//-----------------------------------------------------------------------------
//  
//-----------------------------------------------------------------------------
// Paramtres : 
// - une carte
// Valeur de retour : une liste
//-----------------------------------------------------------------------------


int CartaGene::GetOrdMap(int nbmap, intPtr *vm) const
{

  *vm = new int[NbMS+1];
  // pour compter avoir la taille du groupe directement dans le vecteur.
  (*vm)[NbMS] = 0;
  
  if (Heap.HeapSize == 0) 
    {
      sprintf(bouf,  "Error : Empty heap.\n");
      perr(bouf);
      return -1;
    }
  
  if ( nbmap < 0 || nbmap >= Heap.HeapSize )
    {
      sprintf(bouf,  "Error : This map does not exist.\n");
      perr(bouf);
      return -1;
    }
  
  // pour stocker.
  for (int i = 0; i < NbMS; i++)
    (*vm)[i] = Heap.MapFromId(nbmap)->ordre[i]; 
  
  return NbMS;
}

//-----------------------------------------------------------------------------
//  d'une carte
// Seuls les marqueurs du gnotype sont affichs
//-----------------------------------------------------------------------------
// Paramtres : 
// - l'unit
// - une carte
// Valeur de retour : une liste
//-----------------------------------------------------------------------------

char **** CartaGene::GetHeap(char unit[2],
			     int nbmap) const
{
  
  // pour retourner une liste vide(voir conventions)
  //char * vide = new char[1];
  //vide[0] = '\0';

  switch(unit[0]) 
    {
    case 'h': break;
    case 'k': break;
    default: 
      sprintf(bouf,  "Error : UnitFlag should be 'k' or 'h'.\n");
      perr(bouf);
      return NULL;
    }

  if (Heap.HeapSize == 0) 
    {
      sprintf(bouf,  "Error : Empty heap.\n");
      perr(bouf);
      return NULL;
    }
  
  if ( nbmap == 0 || nbmap > Heap.HeapSize )
    nbmap = Heap.HeapSize;

  // allocation
  
  char**** lmn =  new char***[nbmap + 1];
  lmn[nbmap] = NULL;

  int *tabid = Heap.IdSorted();

  for (int i = 0; i < nbmap; i++)
      lmn[i] = GetMap(unit, tabid[i]);

  delete tabid;

  return(lmn);
}

//-----------------------------------------------------------------------------
//  Affichage dtaill d'une carte
// Seuls les marqueurs du gnotype sont affichs
//-----------------------------------------------------------------------------
// Paramtres : 
// - une carte
// Valeur de retour : 
//-----------------------------------------------------------------------------

void CartaGene::PrintDMap(int nbmap, int envers) const
{

  if (Heap.HeapSize == 0) 
    {
      sprintf(bouf,  "Error : Empty heap.\n");
      perr(bouf);
      return;
    }
  
  if ( nbmap < 0 || nbmap >= Heap.HeapSize )
    {
      sprintf(bouf,  "Error : This map does not exist.\n");
      perr(bouf);
      return;
    }

  sprintf(bouf, "\nMap %2d : log10-likelihood = %8.2f, log-e-likelihood = %8.2f\n", 
	 (Heap.MapFromId(nbmap))->Id, 
	 (Heap.MapFromId(nbmap))->coutEM, 
	 (Heap.MapFromId(nbmap))->coutEM * log(10));
  pout(bouf);
  sprintf(bouf,  "-------:\n");
  pout(bouf);
  ArbreJeu->PrintDMap(Heap.MapFromId(nbmap), envers, Heap.MapFromId(nbmap));
}


//-----------------------------------------------------------------------------
// Rcupration d''un groupe sous la forme d''un vecteur d''indice.
//-----------------------------------------------------------------------------
// Paramtres : 
// - un indice de groupe
// - pointeur sur un vecteur d'entier.
// Valeur de retour : le nombre de marqueurs dans le groupe.
//-----------------------------------------------------------------------------

int CartaGene::GetGroupeI(int idg, intPtr *vm) const
{
  int i, nbg, nbm;
  NodintPtr padj, padjw;
  NodptrPtr plink;

  nbg = nbm = 0;

  if (Group == NULL)
    {
      *vm =  new int[1];
      (*vm)[0] = 0;
      return 0;
    }

  // trouver le groupe.
  plink = Group;

  while (plink != NULL && nbg++ < idg - 1)
    plink = plink->next;
  
  // premier passage pour compter.
  padj = plink->first;
  padjw = padj;
  while (padj != NULL) 
    {
      nbm++;
      padj = padj->next;
    }

  *vm = new int[nbm+1];
  // pour compter avoir la taille du groupe directement dans le vecteur.
  (*vm)[nbm] = 0;

  // deuxime passage pour stocker.
  i = 0;
  padj = padjw;
  while (padj != NULL) 
    {
      (*vm)[i++] = padj->vertex;
      padj = padj->next;
    }

  return nbm;
}

//-----------------------------------------------------------------------------
// Construction d'une carte, calcul et insertion dans le tas
//-----------------------------------------------------------------------------
// Paramtres : 
// Valeur de retour :
//-----------------------------------------------------------------------------

void CartaGene::SinglEM()
{

  // vrification de base

  if (NbMS == 0) 
    {
      sprintf(bouf,  "Error : Empty selection of loci.\n");
      perr(bouf);
      return;
    }

  // vrification de la prsence de tous les marqueurs 
  // dans le BioJeu principal = solution un peu rigide.

  for (int i = 0; i < NbMS; i++)
    if ((BitJeuMarq[MarkSelect[i]] & ArbreJeu->BitJeu) == 0)
      {
	sprintf(bouf, "Error : Unknown Loci Id %d in the main data set %d.\n",
		MarkSelect[i],
		ArbreJeu->Id);
	perr(bouf);
	return;
      }

  Carte LaCarte(this, NbMS, MarkSelect);

  ComputeEM(&LaCarte);
  Heap.Insert(&LaCarte,0);

  if (!QuietFlag) 
    if (VerboseFlag>1) PrintDMap(&LaCarte, 0, &LaCarte);
    else PrintMap(&LaCarte);
}

//-----------------------------------------------------------------------------
// Pour une nouvelle carte pas trop bete.
//-----------------------------------------------------------------------------

void CartaGene::BuildNiceMap(void)
{

  // vrification de base
  
  if (NbMS == 0) 
    {
      sprintf(bouf,  "Error : Empty selection of loci.\n");
      perr(bouf);
      return;
    }
  
  Carte TheMap(this, NbMS, MarkSelect);

  TheMap.BuildNiceMap();

  ComputeEM(&TheMap);
  Heap.Insert(&TheMap,0);

  if (!QuietFlag) 
    if (VerboseFlag>1) PrintDMap(&TheMap, 0, &TheMap);
    else PrintMap(&TheMap);
}
void CartaGene::BuildNiceMapMultiFragment(void)
{

  // vrification de base
  
  if (NbMS == 0) 
    {
      sprintf(bouf,  "Error : Empty selection of loci.\n");
      perr(bouf);
      return;
    }
  
  Carte TheMap(this, NbMS, MarkSelect);

  TheMap.BuildNiceMapMultiFragment();

  ComputeEM(&TheMap);
  Heap.Insert(&TheMap,0);

  if (!QuietFlag) 
    if (VerboseFlag>1) PrintDMap(&TheMap, 0, &TheMap);
    else PrintMap(&TheMap);
}

//-----------------------------------------------------------------------------
// Pour une nouvelle carte pas trop bete.
//-----------------------------------------------------------------------------

void CartaGene::BuildNiceMapL(void)
{

  // vrification de base
  
  if (NbMS == 0) 
    {
      sprintf(bouf,  "Error : Empty selection of loci.\n");
      perr(bouf);
      return;
    }
  
  Carte TheMap(this, NbMS, MarkSelect);

  TheMap.BuildNiceMapL();

  ComputeEM(&TheMap);
  Heap.Insert(&TheMap,0);

  if (!QuietFlag) 
    if (VerboseFlag>1) PrintDMap(&TheMap, 0, &TheMap);
    else PrintMap(&TheMap);
}
void CartaGene::BuildNiceMapLMultiFragment(void)
{

  // vrification de base
  
  if (NbMS == 0) 
    {
      sprintf(bouf,  "Error : Empty selection of loci.\n");
      perr(bouf);
      return;
    }
  
  Carte TheMap(this, NbMS, MarkSelect);

  TheMap.BuildNiceMapLMultiFragment();

  ComputeEM(&TheMap);
  Heap.Insert(&TheMap,0);

  if (!QuietFlag) 
    if (VerboseFlag>1) PrintDMap(&TheMap, 0, &TheMap);
    else PrintMap(&TheMap);
}

//-----------------------------------------------------------------------------
// Post traitement: Pour chaque marqueur,
// compare la probabilit courante avec la probabilit obtenue en
// dplassant le marqueur vers un autre intervalle.  Ne met plus  jour
// avec le meilleur lment (s'il y en a un). Affecte seulement le TAS
// Affiche la matrice des diffrences.
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

void CartaGene::Polish()
{

  if (Heap.HeapSize == 0) 
    {
      sprintf(bouf,  "Error : Empty heap.\n");
      perr(bouf);
      return;
    }
  
  Carte *InitMap;

  Carte TheMap(this, NbMS, MarkSelect);

  InitMap = Heap.Best();
  InitMap->CopyFMap(&TheMap);

  TheMap.Polish();
}
//-----------------------------------------------------------------------------
// Post traitement: Pour chaque marqueur,
// compare la probabilit courante avec la probabilit obtenue en
// dplassant le marqueur vers un autre intervalle.  Ne met plus  jour
// avec le meilleur lment (s'il y en a un). N'Affectepas le TAS
// Affiche la matrice des diffrences.
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

void CartaGene::Polishtest(int *vmi, int nbmi)
{
  int i=0,j=0;
  // une petite vrif...
  for (i = 0; i < nbmi; i++)
  {

    if (vmi[i] > NbMarqueur || vmi[i]< 1) 
      {
      fprintf(stderr, "Error : Unknown Loci Id %d.\n", vmi[i]);
      return;
      }
    
    for (j = i+1; j< nbmi; j++) 
      if (vmi[i] == vmi[j])
	{
	  fprintf(stderr, "Error : Double occurrence Loci Id %d.\n", vmi[i]);
	  return;
	}
  }
  if (Heap.HeapSize == 0) 
    {
      fprintf(stderr, "Error : Empty heap.\n");
      return;
    }
  fflush(stdout);
   
  Carte *InitMap;

  Carte TheMap(this, nbmi, vmi);
  ComputeEM(&TheMap);
  //  InitMap = Heap.Best();
  //  InitMap->CopyFMap(&TheMap);
  fflush(stdout);
 
  TheMap.Polishtest(nbmi);
}
//-----------------------------------------------------------------------------
// les flips
//-----------------------------------------------------------------------------
// Paramtres : 
// - fentre
// - seuil
// - itratif
// Valeur de retour : 
//-----------------------------------------------------------------------------


void CartaGene::Flips(int n, double thres, int iter)
{

  // quelques vrif
  
  if (Heap.HeapSize == 0) 
    {
      sprintf(bouf,  "Error : Empty heap.\n");
      perr(bouf);
      return;
    }
  
  if (n < 2 || n > 9)
    {
      sprintf(bouf,  
	      "Error : The size of the window should be between 2 an 9.\n");
      perr(bouf);
      return;
    }  

  if (iter != 0 && iter != 1)
    {
      sprintf(bouf,  
	      "Error : The flag for iteration should be 0 or 1.\n");
      perr(bouf);
      return;
    }  
   
  Carte *InitMap;

  Carte TheMap(this, NbMS, MarkSelect);

  InitMap = Heap.Best();
  InitMap->CopyFMap(&TheMap);

  TheMap.Flips(n, thres, iter);

}
//-----------------------------------------------------------------------------
// Redimensionnement du tas
// L'ancien est recopi dans le nouveau.
//-----------------------------------------------------------------------------
// Paramtres : 
// - la taille du nouveau tas.
//-----------------------------------------------------------------------------

void CartaGene::ResizeHeap(int size)
{

  if (size < 1)
    {
      sprintf(bouf,  
	      "Error : The size of the heap should be greater than 0.\n");
      perr(bouf);
      return;
    }  
  
  Tas *newheap = new Tas();
  Carte *carte;
  
  newheap->EquivalenceFlag = Heap.EquivalenceFlag;

  newheap->Init(this, size);

  // on met temporairement la robustness a 0 pour eviter la sortie
  // precoce.
  double tmpRobustness = Robustness;
  Robustness = 1e111;

  while (Heap.HeapSize)
    {
      carte = Heap.Worst();
      // on deconverge la carte pour reevaluation a l'insertion.
      carte->UnConverge();
      newheap->Insert(carte, 0);
      Heap.Extract();
    }
  Robustness = tmpRobustness;

  Heap = *newheap;
}

//-----------------------------------------------------------------------------
// Fusion des listes des marqueurs que representent deux marqueurs 
//-----------------------------------------------------------------------------
// Paramtres : 
// - les deux marqueurs
//-----------------------------------------------------------------------------
void CartaGene::MergeRepresents(int m1, int m2)
{
    int i = m1;
    
    while (Represents[i] != 0) i = Represents[i];
    Represents[i] = m2;
}
//-----------------------------------------------------------------------------
// Fusion de deux marqueurs
//-----------------------------------------------------------------------------
// Paramtres : 
// - les deux marqueurs
// Retourne 0 si Ok 1 si probleme compatibilite.
//-----------------------------------------------------------------------------
int CartaGene::Merge(int m1, int m2)
{
    if (m1 > NbMarqueur || m1 < 1) {
	sprintf(bouf, "Error : Unknown Locus %d.\n", m1);
	perr(bouf);
	return 1;
    }

    if (m2 > NbMarqueur || m2 < 1) {
	sprintf(bouf, "Error : Unknown Locus %d.\n", m2);
	perr(bouf);
	return 1;
    }

    if (m1 == m2)  {
	sprintf(bouf,  "Error : Cannot merge a locus with itself\n");
	perr(bouf);
	return 1;
    }

    // On verifie la compatibilite
    if (ArbreJeu->Compatible(m1,m2)) {
	ArbreJeu->Merge(m1,m2);
	Merged[m2] = 1;
	MergeRepresents(m1,m2);
	sprintf(bouf, "Markers %d and %d merged in %d.\n",m1,m2,m1);
	pout(bouf);
	
	// on efface le marqueur source de la selection s'il en fait partie
	int j = 0;

	for (int i = 0; i < NbMS; i++) 
	  if (MarkSelect[i] == m2) continue;
	  else
	    MarkSelect[j++] = MarkSelect[i]; 
	
	if (j == NbMS - 1)
	  {
	    NbMS--;
	    Heap.Init(this, Heap.MaxHeapSize);
	  }
	    //	ArbreJeu->ComputeTwoPoints();
	return 0;
    }
    else {
	sprintf(bouf,  "The two markers are incompatibles.\n");
	pout(bouf);
	return 1;
    }
}
//-----------------------------------------------------------------------------
// return a list of list of merged loci by id,
// The first locus of the sublist represent the others that are following.
//-----------------------------------------------------------------------------
// Parameters : 
//-----------------------------------------------------------------------------

char ** CartaGene::GetMerged()
{
  
  char* lineres;
  int j,k;

  // allocation
  
  // majore
  char** listm =  new char*[NbMarqueur + 1];
  //convention
  listm[NbMarqueur] = NULL;
  
  k = 0;
  
  for (int i = 1; i <= NbMarqueur; i++) 
    {
      j = Represents[i];
      
      if (j != 0 && ! Merged[i]) {
	// majore
	lineres = new char[(NbMarqueur + 1)*4];
	sprintf(lineres, "%d ",i);
	while (j != 0) {
	  sprintf(lineres, "%s%d ",lineres, j);
	  j = Represents[j];
	}
	listm[k++] = lineres;    
      }
       
    }
  //convention
  listm[k] = NULL; 
  return (listm);
}

//-----------------------------------------------------------------------------
// Affectation de la liste de marqueurs slectionns
//-----------------------------------------------------------------------------
// Paramtres : 
// - vecteur et nombres de marqueurs.
//-----------------------------------------------------------------------------

void CartaGene::ChangeSel(int *vm, int nbm) 
{

  int i,j;

  // une petite vrif...
  for (i = 0; i < nbm; i++) {
    
    if (vm[i] > NbMarqueur || vm[i]< 1) 
      {
      sprintf(bouf, "Error : Unknown Loci Id %d.\n", vm[i]);
      perr(bouf);
      return;
      }
    
    for (j = i+1; j< nbm; j++) 
      if (vm[i] == vm[j])
	{
	  sprintf(bouf, "Error : Double occurrence Loci Id %d.\n", vm[i]);
	  perr(bouf);
	  return;
	}
  }

  // not optimized
  int found, same = (nbm  == NbMS);
  
  if (same)
    for (i = 0; i < nbm; i++) {
      found = 0;
      for (j = 0; j < nbm; j++)
	if (MarkSelect[j] == vm[i]) {
	  found = 1;
	  break;
	}
      if (!found) { same = 0; break; }
    }

  if (!same) {
    delete MarkSelect; 
    MarkSelect = new int[nbm];
    Heap.Init(this, Heap.MaxHeapSize);
  }
  
  for (i = 0; i < nbm; i++) MarkSelect[i] = vm[i]; 
  NbMS = nbm; 
  
}

//-----------------------------------------------------------------------------
// Rcupration de la slection de M. sous la forme d''un vecteur d''indice.
//-----------------------------------------------------------------------------
// Paramtres : 
// - pointeur sur un vecteur d'entier.
// Valeur de retour : le nombre de marqueurs dans le groupe.
//-----------------------------------------------------------------------------

int CartaGene::GetSel(intPtr *vm) const
{
  *vm = new int[NbMS+1];
  // pour compter avoir la taille du groupe directement dans le vecteur.
  (*vm)[NbMS] = 0;

  // pour stocker.
  for (int i = 0; i < NbMS; i++)
    (*vm)[i] = MarkSelect[i]; 

  return NbMS;
}
//-----------------------------------------------------------------------------
// Rcupration de tous les M. sous la forme d''un vecteur d''indice.
//-----------------------------------------------------------------------------
// Paramtres : 
// - pointeur sur un vecteur d'entier.
// Valeur de retour : le nombre de marqueurs dans le groupe.
//-----------------------------------------------------------------------------

int CartaGene::GetAll(intPtr *vm) const
{
  *vm = new int[NbMarqueur+1];
  // pour compter avoir la taille du groupe directement dans le vecteur.
  (*vm)[NbMarqueur] = 0;

  // pour stocker.
  for (int i = 0; i < NbMarqueur; i++)
    (*vm)[i] = i+1; 

  return NbMarqueur;
}
//-----------------------------------------------------------------------------
// Fixer les tolerances de convergence EM
//-----------------------------------------------------------------------------
// Paramtres : 
// - le niveau de convergence fin
// - le niveau de convergence grossier
// Valeur de retour : none
//-----------------------------------------------------------------------------
void CartaGene::SetTolerance(double Fin,double Gros)
{
  if (Fin <= 0.0) {
    sprintf(bouf,  "Error : fine convergence threshold must be > 0\n");
    perr(bouf);
    return;
  } 

  if (Gros <= 0.0) {
    sprintf(bouf,  "Error : raw convergence threshold must be > 0\n");
    perr(bouf);
    return;
  }
  
  if (Gros < Fin) {
    sprintf(bouf,  "raw convergence threshold must be > fine one\n");
    perr(bouf);
    return;
  } 

  ArbreJeu->Epsilon2 = Fin;
  ArbreJeu->Epsilon1 = Gros;

  // on resize le tas a la meme taille pour reevaluer les cartes du
  // tas.
  if (Heap.HeapSize >0)
    ResizeHeap(Heap.HeapSize);
}
//-----------------------------------------------------------------------------
// Fixer le seuil de fiabilit des cartes frameworks
//-----------------------------------------------------------------------------
// Paramtres : 
// - l'cart de vraisemblance relatif  la meilleure carte (default = -3.)
// Valeur de retour : none
//-----------------------------------------------------------------------------
void CartaGene::SetRobustness(double robust)
{
  Robustness = robust;

  // verification de la robustesse de la meilleure carte courante
  if (Heap.HeapSize >= 2 && Heap.Delta() + Robustness < 0.) {
    sprintf(bouf,  "Current best map is not robust!\n");
    perr(bouf);
  }
}
//-----------------------------------------------------------------------------
// Affectation de la verbosit
//-----------------------------------------------------------------------------
// Paramtres : 
// - le niveau
// Valeur de retour : 
//-----------------------------------------------------------------------------

void CartaGene::SetVerbose(int flag)
{
   if (flag < 0 || flag > 2) 
    {
      sprintf(bouf,  "Error : possible values are 0, 1 or 2.\n");
      perr(bouf);
      return;
    }

  VerboseFlag = flag;
}
//-----------------------------------------------------------------------------
// Affectation du silencieux
//-----------------------------------------------------------------------------
// Paramtres : 
// - le niveau
// Valeur de retour : 
//-----------------------------------------------------------------------------

void CartaGene::SetQuiet(int flag)
{
   if (flag < 0 || flag > 1) 
    {
      sprintf(bouf,  "Error : posible values are 0 or 1.\n");
      perr(bouf);
      return;
    }

  QuietFlag = flag;
}

//-----------------------------------------------------------------------------
// Algorithme genetique.
//-----------------------------------------------------------------------------
// Paramtres : 
// - ...
// Valeur de retour : 
//-----------------------------------------------------------------------------

void CartaGene::AlgoGen( int nb_gens,
			 int nb_elements,
			 int selection_number,
			 float pcross,
			 float pmut,
			 int evolutive_fitness)
{
  // quelques vrifications 

  if (Heap.HeapSize == 0 && nb_elements == 0) 
    {
      sprintf(bouf,  "Error : Empty heap.\n");
      perr(bouf);
      return;
    }
  
  if (nb_gens < 0)
    {
      sprintf(bouf,  "Error : value expected for the number of generation :  >= 0.\n");
      perr(bouf);
      return;
    }
  if (nb_elements < 0)
    {
      sprintf(bouf,  "Error : value expected for the size of the population: >= 0.\n");
      perr(bouf);
      return;
    }
  if (nb_elements > NbMS)
    {
      sprintf(bouf,  "Error : value expected for the size of the population <= number of loci.\n");
      perr(bouf);
      return;
    }

  if (selection_number < 0)
    {
      sprintf(bouf,  "Error : value expected for the type of the selection : >= 0.\n");
      perr(bouf);
      return;
    }
  if ((pcross < 0.0) || (pcross > 1.0))
    {
      sprintf(bouf,  "Error : value expected for the probability of crossing over : 0.0 < & < 1.0.\n");
      perr(bouf);
      return;
    }
  if ((pmut < 0.0) || (pmut > 1.0))
    {
      sprintf(bouf,  "Error : value expected for the probability of mutation : 0.0 < & < 1.0.\n");
      perr(bouf);
      return;
    }
  if (evolutive_fitness < 0)
    {
      sprintf(bouf,  "Error : value expected for the evolutive fitness : >= 0.\n");
      perr(bouf);
      return;
    }

  // fin des vrifications

  Algogen *lalgogen;

  lalgogen = new Algogen(this);

  lalgogen->Run(nb_gens,
		nb_elements,
		selection_number,
		pcross,
		pmut,
		evolutive_fitness);

  delete lalgogen;
  
}
  
//-----------------------------------------------------------------------------
// Retourne le nom d'un marqueur
//-----------------------------------------------------------------------------
// Paramtres : 
// - ...
// Valeur de retour : 
//-----------------------------------------------------------------------------

char * CartaGene::GetMrkName(int mid)
{
  char *vide;
  char *lineres;

  // pour retourner une liste vide (voir conventions)
  vide = new char[1];
  vide[0] = '\0';
  
  if (mid > NbMarqueur || mid < 1) 
    {
      sprintf(bouf, "Error : Unknown Locus %d.\n", mid);
      perr(bouf);
      return vide;
    }

  lineres = new char[strlen(NomMarq[mid]) + 1];

  sprintf(lineres,  NomMarq[mid]);

  return lineres;
  
}

//-----------------------------------------------------------------------------
// Retourne le numro d'un marqueur
//-----------------------------------------------------------------------------
// Paramtres : 
// - ...
// Valeur de retour : 
//-----------------------------------------------------------------------------

int CartaGene::GetMrkId(char *mname)
{
  for (int i = 1; i <= NbMarqueur; i++) 
    if (!strcmp(mname, NomMarq[i])) return(i);    
  
  sprintf(bouf, "Error : Unknown Locus %s.\n", mname);
  perr(bouf);
  return -1; 
  
}

//-----------------------------------------------------------------------------
// Retourne le numro du dernier marqueur
//-----------------------------------------------------------------------------
// Paramtres : 
// - ...
// Valeur de retour : 
//-----------------------------------------------------------------------------

int CartaGene::LastMrk()
{
  return NbMarqueur;  
}

//-----------------------------------------------------------------------------
// Retourne un tableau des distance cumulees 
//-----------------------------------------------------------------------------
// Paramtres : 
// - ...
// Valeur de retour : 
//-----------------------------------------------------------------------------

void CartaGene::CalculDist(double *distd, Carte *data, int NbMS, int NbMF)
{
   int i,k,itr;
   double d, totald = 0.0;

   distd[0] = 0;
   for (k = 0; k < data->NbMarqueur; k++)
   {

      i = k;
      itr = i;

   
      if (k < data->NbMarqueur-1)
	(ArbreJeu->HasRH() ? d = Theta2Ray(data->tr[itr])*100 : d = Haldane(data->tr[itr])*100);
      totald += d;
      distd[k+1] = totald;
  } // for (k=
}


