//-----------------------------------------------------------------------------
//                            CarthaGene
//
// Martin Bouchez, Patrick Chabrier, Christine Gaspin and Thomas Schiex. 
// Unite de Biometrie et Intelligence Artificielle, Toulouse, France
//
// Copyright 1997 INRA (Institut National de la Recherche Agronomique).
// This file is distributed under the terms of the Q Public License version 1.0.
// 
// $Id: CartaGene.h,v 1.35 2003/06/26 16:02:39 degivry Exp $
//
// Description : Prototype de l''API du module.
// Divers : 
//-----------------------------------------------------------------------------

#ifndef _CARTAGENE_H
#define _CARTAGENE_H

#include <ctype.h>
#include "CGtypes.h"
#include "Carte.h"
#include "Tas.h"

#ifdef ihm
#include <tcl.h>
#define fflush(stdout) fflush(stdout); while (Tcl_DoOneEvent(TCL_DONT_WAIT))
#endif

#define MAXNBJEU 32
#define MAXNBMARKERS 10000

class BioJeu;


/** Interface du noyau de calcul */
class CartaGene {
 public:
  
  /** Version du produit */ 
  char LaVersion[128];
  
  /** Nombre total de marqueurs */
  int NbMarqueur;
  
  /** Nombre total de jeux */
  int NbJeu;

  /** Nombre maximum de jeux chargeable simultanement */
  /* Cette limitation est imposee par le type utilise pour BitJeuMarq */
  /*static const int MAXNBJEU = 32;*/

  /** Nombre total d'individus. */
  int NbIndividu;
  
  /** Nom de chaque marqueur. */
  charPtr *NomMarq;
  
  /** Le marqueur est la fusion de qui */
  intPtr Represents;

  /** Le marqueur a-t-il ete fusionne ? */
  charPtr Merged;

  /** Champs de bits de chaque marqueur "i". */
  /* BitJeuMarq[i] a le jeme bit positionne a 1 ssi le marqueur i est
     defini dans le jeu de donnees "j" */
  int *BitJeuMarq;
  
  /** Tableau des jeux. */
  BioJeu **Jeu;
  
  /** Racine de l'arbre des jeux. */
  BioJeu *ArbreJeu;
  
  /** La liste des groupes(dernire dtermination). */
  NodptrPtr Group;
  
  /** Le seuil de distance utilis pour le calcul des groupes. */
  double DisThres;
  
  /** Le seuil de LOD utilis pour le calcul des groupes. */
  double LODThres;

  /** Le seuil de fiabilit des cartes frameworks. 
      utilis pour la vrification de cartes frameworks.
      arrte la recherche lorsque L(SecondBestMap) > L(BestMap) + Robustness */
  double Robustness;

  /** Drapeau de demande d'arrt d'une commande*/
  int StopFlag;

  /** Drapeau du mode verbeux des rsultats en cours de traitement. */
  int VerboseFlag;
  
  /** Drapeau du mode silencieux sur la progression du traitement. */
  int QuietFlag;
  
  /** Structure de stockage des cartes */
  Tas Heap;

  /** le vecteur des marqueurs actifs pour la recherche */
  int *MarkSelect;

  /** le nombre des marqueurs actifs */
  int NbMS;

  /** Constructeur. */
  CartaGene();
  
  /** Destructeur. */
  ~CartaGene();

  /** Demande de trace.
      @param fileName le nom du fichier
    */
  void Tracer(char *fileName);

  /** Affiche la version*/
  void Version(void);
  
  /** Efface toutes les informations*/
  void Clear(void);

  /** Entre la clef de license */
  void License(const char *key);
  
  /** Chargement d'un jeu de donnes.
      @param fileName le nom du fichier
      @return un message d'information sur le jeu de donnes charg
    */
  char * CharJeu(char *fileName);
  
  /** Affichage de tous les marqueurs. */
  void AffMarq(void);
  
  /** Affichage de tous les jeux. */
  void AffJeu(void);

  /** retourne de l'info sur tous les jeux. */
  char** GetArbre(void);
  
  /** Affichage de l'chantillon d'un jeu.
      @param nbjeu le numro du jeu
  */
  void DumpEch(int nbjeu);
  
  /** Affichage des LOD 2 points d'un jeu.
      @param nbjeu le numro du jeu
  */
  void DumpTwoPointsLOD(int nbjeu);

  /** Affichage des LOD 2 points des marqueurs de la slection
  */
  void PrintTwoPointsLOD(void);

  /** Affichage des marqueurs ayant un haplotype compatible
  */
  void DumpDouble(void);

  /** Affichage des FR 2 points d'un jeu.
  */
  void PrintTwoPointsFR(void);

  /** Affichage des FR 2 points d'un jeu.
  */
  void DumpTwoPointsFR(void);
  
  /** Accs  une valeur LOD score 2pt.
      @param m1 le numro du premier marqueur
      @param m2 le numro du second marqueur
      @return le LOD
  */
  double GetTwoPointsLOD(int m1, int m2) const;
  
  /** Accs  la fraction de recombinaison 2pt.sdist
      @param m1 le numro du premier marqueur
      @param m2 le numro du second marqueur
      @return la fraction de recombinant
  */
  double GetTwoPointsFR(int m1, int m2) const;
  
  /** Accs  la distance 2pt.
      @param m1 le numro du premier marqueur
      @param m2 le numro du second marqueur
      @return la fraction de recombinant
  */
  double GetTwoPointsDH(int m1, int m2) const;

  /** Affichage des LOD 2 points des marqueurs de la slection
        @param unit 1 c. pou spcifier l'unit
  */
  void PrintTwoPointsDist(char unit[2]);

  /** Calcul des groupes d'un jeu.
      @param disthres le seuil de distance
      @param lodthres le seuil de lod
      @return le nombre de groupes trouv
  */
  int Groupe(double disthres, double lodthres);

  /** Rcupration d'un groupe sous la forme d'un vecteur d'indice.
      @param idg le numro du groupe
      @param le vecteur de numro de marqueurs
      @return le nombre de marqueur dans le groupe
  */
  int GetGroupeI(int idg, intPtr *vm) const;
  
  /** Fusion gntique entre deux jeux.
      @param gauche le numro du premier jeu
      @param droite le numro du second jeu
      @return un message d'information sur le jeu de donnes fusionn
  */
  char *MerGen(int gauche, int droite);
  
  /** Fusion sur l'ordre entre deux jeux.
      @param gauche le numro du premier jeu
      @param droite le numro du second jeu
      @return un message d'information sur le jeu de donnes fusionn
  */
  char *MergOr(int gauche, int droite);
  
  /** Calcul de la vraisemblance maximum par EM.
      @param map la carte
      @return le maximum de vraisemblance
  */
  double ComputeEM(Carte *map);

  /** Calcul de la vraisemblance maximum par EM.
      Une interface plus sophistique du calcul de la vraisemblance par EM:
      On converge jusqu' epsilon1. Si la vraisemblance est infrieure 
      threshold on laisse tomber, sinon on converge  epsilon2.
      @param data la carte
      @param threshold niveau a partir duquel on convergera a epsilon2
      @return le maximum de vraissemblance
  */
  double ComputeEMS (Carte *data, 
		     double threshold);
  
  /** Construction d'une carte, calcul et insertion dans le tas
    */
  void SinglEM();

  /** retourne une carte sous forme de liste
      @param unit l'unit de distance: k = kosambi ou centiray, h = haldane ou centiray.
      @param nbmap le numro de la carte dans le tas
      @return une liste contenant le nombre d'ordres, les nombres de marqueurs par ordre, les LOD par odre, chaque marqueur et sa position. 
  */
  char *** GetMap(char unit[2],int nbmap) const;

  /** retourne les cartes sous les formes d'une liste.
      @param nbmap le numro d'une carte dans le tas
      @return une liste contenant les indices de locus
  */
  int GetOrdMap(int nbmap, intPtr *vm) const;

  /** retourne les cartes sous les formes d'une liste.
      @param unit l'unit de distance: k = kosambi ou centiray, h = haldane ou centiray.
      @param nbmap les nbmap meilleurs cartes.
      @return une liste contenant le nombre d'ordres, les nombres de marqueurs par ordre, les LOD par odre, chaque marqueur et sa position. Pour chaque carte 
  */
  char **** GetHeap(char unit[2],int nbmap) const;

  /** Affichage simple d'une carte
      @param data la carte
  */
  void PrintMap(Carte *data) const;

  /** Affichage dtaill d'une carte
      @param data la carte
      @param envers  l'envers ou  l'endroit
      @param dataref la carte de rfrence
  */
  void PrintDMap(Carte *data, int envers, Carte *dataref) const;

  /** Affichage simple d'une carte
      @param nbmap le numro de la carte dans le tas
  */
  void PrintMap(int nbmap) const;

  /** Affichage dtaill d'une carte
      @param nbmap le numro de la carte dans le tas
      @param envers  l'envers ou  l'endroit
  */
  void PrintDMap(int nbmap, int envers) const;

  /** Affichage des deltas entre carte FW et carte FW + marqueur
      @param structure de resultat
      @param Carte Framework
      @param le nombre de marqueur de la carte FW
      @param le nmbre de marqueur a tester/carte FW
      @param le seuil pour afficher le delta
      @param le meilleur intervalle pour inserer le marqueur
  */
  void PrintVMap(struct testmartin *pstruct, Carte *data, double *distd, int nbm, int nbtas, double AM);
  
  /** Fixation des seuils de convergence EM
      @param Fin le seuild de convergence fine
      @param Gros le seuild de convergence grossiere
  */
  void SetTolerance(double Fin, double Gros);
  
  /** Fixation du seuil de fiabilit des cartes frameworks.
      @param cart de vraisemblance relatif  la meilleure carte
  */
  void SetRobustness(double robust);

  /** Affectation de la verbosit
      @param flag le niveau
  */
  void SetVerbose(int flag);

  /** Affectation du silencieux
      @param flag le niveau
  */
  void SetQuiet(int flag);

  /** Pour une nouvelle carte pas trop bte.*/
  void BuildNiceMap(void);
  void BuildNiceMapMultiFragment(void);

  /** Pour une nouvelle carte pas trop bte.*/
  void BuildNiceMapL(void);
  void BuildNiceMapLMultiFragment(void);

  /** Build
 
      @param NC le nombre de cartes construites simultanments
  */
  void Build(int NC);

  /** Build

      @param SM Delta minimum ncessaire  l'acceptation de l'insertion d'un marqueur
      @param  AM Seuil  l'intrieur duquel les ordres sont conservs.
      @param vm vecteur d'indice de marqueurs
      @param nombre de marqueurs dans la liste.
      @param Demande test marqueurs non FW.
  */
  void BuildFW(double SM, double AM, int *vmi, int nbmi, int DM);

  /** Recuit Simul
      La recherche de la meilleure carte est faite avec l'algorithme de Recuit.
      Durant la recherche, les cartes trouves sont stockes dans le tas et
      ranges dans l'ordre de cout dcroissant.
      @param Tries Nombre d'essai par temprature.
      @param Tinit Temprature de dbut.
      @param Tfinal Temprature finale.
      @param Cooling Vitesse de refroidissement.
  */
  void Annealing(int Tries,
		 double Tinit,
		 double Tfinal,
		 double Cooling);

  /** Lancement de Greedy.
      @param NR Nombre de boucles principales.
      @param NI Nombre d'itrations.
      @param TMin TabooMin.
      @param TMax TabooMax.
  */
  void Greedy(int NR, 
	      int NI,
	      int TMin, 
	      int TMax,
	      int Ratio);
  void GreedyOld(int NR, 
		 int NI,
		 int TMin, 
		 int TMax);
  void GreedyNew(int NR, 
		 int NI,
		 int TMin, 
		 int TMax,
		 int Ratio);

  /** Algorithme genetique.
      @param nb_gen :Nombre de generations (integer)
      @param nb_elements :Nombre d'elements dans la population (integer)
      @param selection_number:# Selection type :(integer)
      #	0 -> roulette wheel
      #	1 -> stochastic remainder without replacement
      @param pcross: Probabilities of crossover and mutation (float)
      @param pmut: Probabilities of crossover and mutation (float)
      @param evolutive_fitness:Evolutive Fitness (integer) 
      (the fitness of the same element changes with the generation number)
      
  */
  void AlgoGen( int nb_gens,
		int nb_elements,
		int selection_number,
		float pcross,
		float pmut,
		int evolutive_fitness);
  

  /** Post traitement polish
      Pour chaque marqueur, compare la probabilit courante avec la
      probabilit obtenue en dplaant le marqueur vers un autre
      intervalle. Met  jour avec le meilleur lment (s'il y en a un)
      et l'insert dans le tas. Affiche la matrice des diffrences (en mode verbose).
  */
  void Polish();
  
  /** Post traitement polishtest
      Pour chaque marqueur, compare la probabilit courante avec la
      probabilit obtenue en dplaant le marqueur vers un autre
      intervalle. NE Met PAS  jour avec le meilleur lment (s'il y en a un)
      et NE l'insere PAS dans le tas.
      Affiche la matrice des diffrences (en mode verbose).
      @param le vecteur des numros de marqueurs.
      @param le nombre de marqueurs
  */
  void Polishtest(int *vm, int nbm);
  
  /** Post traitement flips
      @param n taille de lafentre 
      @param thres le seuil
      @param iter flag pour itratif ou non
  */
  void Flips(int n, double thres, int iter);


  /** Redimensionnement du tas
      L'ancien est recopi dans le nouveau.
      @param size la taille du nouveau tas.
  */
  void ResizeHeap(int size);

  /** Fusion de deux marqueurs
      @param les numros de marqueurs.
  */

  int Merge(int m1, int m2);

  /** return a list of list of merged loci by id,
      The first locus of the sublist represent the others that are following.
  */
  char ** GetMerged();

  /** Changement de la liste des marqueurs exploits
      @param le vecteur des numros de marqueurs.
      @param le nombre de marqueurs
  */
  void ChangeSel(int *vm, int nbm);

  
  /** Rcupration de la slection sous la forme d'un vecteur d'indice.
      @param le vecteur de numro de marqueurs
      @return le nombre de marqueur dans le groupe
  */
  int GetSel(intPtr *vm) const;
  
  /** Rcupration d'intgralit des loci sous la forme d'un vecteur d'indice.
      @param le vecteur de numro de marqueurs
      @return le nombre de marqueur dans le groupe
  */
  int GetAll(intPtr *vm) const;

  /** Retourne le nom d'un marqueur
      @param mid le numro d'un marqueur
      @return le nom du marqueur
  */
  char * GetMrkName(int mid);
  
  /** Retourne le numro d'un marqueur
      @param mname le nom d'un marqueur
      @return le numro du marqueur
  */
  int GetMrkId(char *mname);
  
  /** Retourne le numro du dernier marqueur
      @return le numro du marqueur
  */
  int LastMrk();  
   
  /** Retourne un tableau des distance cumulees 
  */
  void CalculDist(double *distd, Carte *data, int NbMS, int NbMF);

 protected:
  
  /**un ordre intermdiare pour le 3changement
    */
  int *wordre; 
  
 private:
  
  /** Teste le type de donnes d'un fichier.
      @param DataFile le nom du fichier
      @return un type de croisement
    */
  CrossType ProbeCrossType(char DataFile[256]);

  /** Fusionne les listes des marqueurs que representent deux marqueurs
      @param le marqueur 1
      @param le marqueur 2
      @return rien
    */
  void MergeRepresents(int m1, int m2);
  
};

typedef CartaGene *CartaGenePtr;

#endif /* _CARTAGENE_H */
